#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
"""
Functions for clearing pipes.
"""

from __future__ import annotations
from meerschaum.utils.typing import List, SuccessTuple, Any, Optional

def clear(
        action: Optional[List[str]] = None,
        **kw: Any
    ) -> SuccessTuple:
    """
    Clear pipes of their data, or clear the screen.

    Usage:
    - clear
        - Clear the screen.
    - clear pipes --begin 2022-01-01
        - For all pipes, remove rows newer than or equal to `2022-01-01 00:00:00`.

    """
    from meerschaum.actions import choose_subaction
    if not action:
        from meerschaum.utils.formatting._shell import clear_screen
        return clear_screen(kw.get('debug', False)), ''

    options = {
        'pipes': _clear_pipes,
    }
    return choose_subaction(action, options, **kw)


def _clear_pipes(
        action: Optional[List[str]] = None,
        begin: Optional[datetime.datetime] = None,
        end: Optional[datetime.datetime] = None,
        connector_keys: Optional[List[str]] = None,
        metric_keys: Optional[List[str]] = None,
        mrsm_instance: Optional[str] = None,
        location_keys: Optional[List[str]] = None,
        force: bool = False,
        debug: bool = False,
        **kw: Any
    ) -> SuccessTuple:
    """
    Clear pipes' data without dropping any tables.

    """
    from meerschaum import get_pipes
    from meerschaum.utils.formatting import print_tuple
    
    successes = {}
    fails = {}

    pipes = get_pipes(
        as_list=True, connector_keys=connector_keys, metric_keys=metric_keys,
        location_keys=location_keys, mrsm_instance=mrsm_instance, debug=debug, **kw
    )

    if not force:
        if not _ask_with_rowcounts(pipes, begin=begin, end=end, debug=debug, **kw):
            return False, "No rows were deleted."

    for pipe in pipes:
        success, msg = pipe.clear(begin=begin, end=end, debug=debug, **kw)
        print_tuple((success, msg))
        (successes if success else fails)[pipe] = msg

    success = len(successes) > 0
    msg = (
        f"Finished clearing {len(pipes)} pipe" + ('s' if len(pipes) != 1 else '')
        + f'\n    ({len(successes)} succeeded, {len(fails)} failed).'
    )

    return success, msg


def _ask_with_rowcounts(
        pipes: List[meerschaum.Pipe],
        begin: Optional[datetime.datetime] = None,
        end: Optional[datetime.datetime] = None,
        yes: bool = False,
        nopretty: bool = False,
        noask: bool = False,
        debug: bool = False,
        **kw
    ) -> bool:
    """
    Count all of the pipes' rowcounts and confirm with the user that these rows need to be deleted.

    """
    from meerschaum.utils.prompt import yes_no
    from meerschaum.utils.misc import print_options
    from meerschaum.utils.warnings import warn
    for pipe in pipes:
        if not pipe.columns or 'datetime' not in pipe.columns:
            _dt = pipe.guess_datetime()
            is_guess = True
        else:
            _dt = pipe.get_columns('datetime')
            is_guess = False

        if begin is not None or end is not None:
            if is_guess:
                if _dt is None:
                    warn(
                        f"No datetime could be determined for {pipe}!\n"
                        + "    THIS WILL DELETE THE ENTIRE TABLE!",
                        stack = False
                    )
                else:
                    warn(
                        f"A datetime wasn't specified for {pipe}.\n"
                        + f"    Using column \"{_dt}\" for datetime bounds...",
                        stack = False
                    )


    pipes_rowcounts = {p: p.get_rowcount(begin=begin, end=end, debug=debug) for p in pipes} 
    print_options(
        [str(p) + f'\n{rc}\n' for p, rc in pipes_rowcounts.items()],
        header='Number of Rows to be Deleted'
    )
    total_num_rows = sum([rc for p, rc in pipes_rowcounts.items()])
    question = (
        f"Are you sure you want to delete {total_num_rows} rows across {len(pipes)} pipe"
        + ('s' if len(pipes) != 1 else '')
        + " in the following range?\n"
    )
    range_text = (
        (f"\n    Newer than (>=) {begin}" if begin is not None else '')
        + ("\n    Older than (<)" + (' ' if begin else '') + f" {end}" if end is not None else '')
    ) if (begin is not None or end is not None) else '\n    Unbounded (delete all rows)!'
    question += range_text + '\n\n'

    return yes_no(question, yes=yes, nopretty=nopretty, noask=noask, default='n')

### NOTE: This must be the final statement of the module.
###       Any subactions added below these lines will not
###       be added to the `help` docstring.
from meerschaum.actions import choices_docstring as _choices_docstring
clear.__doc__ += _choices_docstring('clear')



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Build the Dash app to be hooked into FastAPI.
"""

from __future__ import annotations
import uuid

from meerschaum.utils.packages import (
    attempt_import, import_dcc, import_html, _monkey_patch_get_distribution
)
flask_compress = attempt_import('flask_compress', lazy=False)
_monkey_patch_get_distribution('flask-compress', flask_compress.__version__)
dash = attempt_import('dash', lazy=False)


from meerschaum.utils.typing import List, Optional
from meerschaum.config.static import _static_config
from meerschaum.api import (
    app as fastapi_app,
    debug,
    _get_pipes,
    get_pipe as get_api_pipe,
    pipes as api_pipes,
    get_api_connector,
    endpoints,
)

from meerschaum.connectors.parse import parse_instance_keys
import warnings
### Suppress the depreciation warnings from importing enrich.
with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    enrich = attempt_import('dash_extensions.enrich', lazy=False)
html, dcc = import_html(), import_dcc()
from meerschaum.api.dash.components import location

active_sessions = {}
authenticated_sessions = {}
unauthenticated_sessions = {}
running_jobs = {}
running_monitors = {}
stopped_jobs = {}
stopped_monitors = {}

stylesheets = [
    '/static/css/bootstrap.min.css',
    '/static/css/dbc_dark.css',
    '/static/css/dash.css',
    #  '/static/js/node_modules/xterm/css/xterm.css',
]
scripts = ['/static/js/node_modules/xterm/lib/xterm.js']
dash_app = enrich.DashProxy(
    __name__,
    title = 'Meerschaum Web',
    requests_pathname_prefix = endpoints['dash'] + '/',
    external_stylesheets = stylesheets,
    update_title = None,
    suppress_callback_exceptions = True,
    transforms = [
        enrich.TriggerTransform(),
        enrich.MultiplexerTransform(),
    ],
)

dash_app.layout = html.Div([
    location,
    dcc.Store(id='session-store', storage_type='session', data={}),
    html.Div([], id='page-layout-div'),
])

import meerschaum.api.dash.pages as pages
import meerschaum.api.dash.callbacks as callbacks

fastapi_middleware_wsgi = attempt_import('fastapi.middleware.wsgi')
fastapi_app.mount(
    endpoints['dash'], fastapi_middleware_wsgi.WSGIMiddleware(dash_app.server)
)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Functions for interacting with the user.
"""

from __future__ import annotations
import os
from meerschaum.utils.typing import Any, Union, Optional, Tuple, List

def prompt(
        question: str,
        icon: bool = True,
        default: Union[str, Tuple[str, str], None] = None,
        detect_password: bool = True,
        is_password: bool = False,
        wrap_lines: bool = True,
        noask: bool = False,
        **kw: Any
    ) -> str:
    """
    Ask the user a question and return the answer.
    Wrapper around `prompt_toolkit.prompt()` with modified behavior.
    For example, an empty string returns default instead of printing it for the user to delete
    (`prompt_toolkit` behavior).

    Parameters
    ----------
    question: str
        The question to print to the user.

    icon: bool, default True
        If True, prepend the configured icon.

    default: Union[str, Tuple[str, str], None], default None
        If the response is '', return the default value.

    detect_password: bool, default True
        If `True`, set the input method to a censored password box if the word `password`
        appears in the question.

    is_password: default False
        If `True`, set the input method to a censored password box.
        May be overridden by `detect_password` unless `detect_password` is set to `False`.

    wrap_lines: bool, default True
        If `True`, wrap the text across multiple lines.
        Flag is passed onto `prompt_toolkit`.

    noask: bool, default False
        If `True`, only print the question and return the default answer.

    Returns
    -------
    A `str` of the input provided by the user.

    """
    from meerschaum.utils.packages import attempt_import
    from meerschaum.utils.formatting import colored, ANSI, CHARSET, highlight_pipes, fill_ansi
    from meerschaum.config import get_config
    from meerschaum.config.static import _static_config
    noask = check_noask(noask)
    if not noask:
        prompt_toolkit = attempt_import('prompt_toolkit')
    question_config = get_config('formatting', 'question', patch=True)

    ### if a default is provided, append it to the question.
    default_answer = default
    if default is not None:
        question += f" (default: "
        if isinstance(default, tuple) and len(default) > 1:
            question += f"{default[0]} [{default[1]}]"
            default_answer = default[0]
        else:
            question += f"{default}"
        question += ")"

    ### detect password
    if (detect_password and 'password' in question.lower()) or is_password:
        kw['is_password'] = True
  
    ### Add the icon and only color the first line.
    lines = question.split('\n')
    first_line = lines[0]
    other_lines = '' if len(lines) <= 1 else '\n'.join(lines[1:])

    if ANSI:
        first_line = fill_ansi(highlight_pipes(first_line), **question_config['ansi']['rich'])
        other_lines = highlight_pipes(other_lines)

    _icon = question_config[CHARSET]['icon']
    question = (' ' + _icon + ' ') if icon and len(_icon) > 0 else ''
    question += first_line
    if len(other_lines) > 0:
        question += '\n' + other_lines
    question += ' '

    answer = (
        prompt_toolkit.prompt(
            prompt_toolkit.formatted_text.ANSI(question),
            wrap_lines = wrap_lines,
            **kw
        ) if not noask else ''
    )
    if noask:
        print(question)
    if answer == '' and default is not None:
        return default_answer
    return answer


def yes_no(
        question: str = '',
        options: Tuple[str, str] = ('y', 'n'),
        default: str = 'y',
        wrappers: Tuple[str, str] = ('[', ']'),
        icon: bool = True,
        yes: bool = False,
        noask: bool = False,
        **kw : Any
    ) -> bool:
    """
    Print a question and prompt the user with a yes / no input.
    Returns `True` for `'yes'`, False for `'no'`.

    Parameters
    ----------
    question: str, default ''
        The question to print to the user.

    options: Tuple[str, str], default ('y', 'n')
        The `y/n` options. The first is considered `True`, and all options must be lower case.

    default: str, default y
        The default option. Is represented with a capital to distinguish that it's the default.

    wrappers: Tuple[str, str], default ('[', ']')
        Text to print around the '[y/n]' options.

    icon: bool, default True
        If True, prepend the configured question icon.

    Returns
    -------
    A bool indicating the user's choice.

    Examples
    --------
    ```python-repl
    >>> yes_no("Do you like me?", default='y')
     ❓ Do you like me? [Y/n]
    True
    >>> yes_no("Cats or dogs?", options=('cats', 'dogs'))
     ❓ Cats or dogs? [cats/dogs]
     Please enter a valid response.
     ❓ Cats or dogs? [cats/dogs] dogs
    False
    ```
    """
    from meerschaum.utils.warnings import error, warn
    from meerschaum.utils.formatting import ANSI, UNICODE
    from meerschaum.utils.packages import attempt_import

    default = options[0] if yes else default
    noask = yes or check_noask(noask)

    ending = f" {wrappers[0]}" + "/".join(
        [
            o.upper() if o.lower() == default.lower()
            else o.lower() for o in options
        ]
    ) + f"{wrappers[1]}"
    while True:
        try:
            answer = prompt(question + ending, icon=icon, detect_password=False, noask=noask)
            success = True
        except KeyboardInterrupt:
            success = False
        
        if not success:
            error(f"Error getting response. Aborting...", stack=False)
        if answer == "":
            answer = default

        if answer.lower() in options:
            break
        warn('Please enter a valid reponse.', stack=False)
    
    return answer.lower() == options[0].lower()

def choose(
        question: str,
        choices: List[str],
        default: Optional[str] = None,
        numeric: bool = True,
        multiple: bool = False,
        delimiter: str = ',',
        icon: bool = True,
        warn: bool = True,
        noask: bool = False,
        **kw
    ) -> Union[str, Tuple[str], None]:
    """
    Present a list of options and return the user's choice.

    Parameters
    ----------
    question: str
        The question to be printed.

    choices: List[str]
        A list of options.

    default: Optional[str], default None
        If the user declines to enter a choice, return this value.

    numeric: bool, default True
        If `True`, number the items in the list and ask for a number as input.
        If `False`, require the user to type the complete string.

    multiple: bool, default False
        If `True`, allow the user to choose multiple answers separated by `delimiter`.

    delimiter: str, default ','
        If `multiple`, separate answers by this string. Raise a warning if this string is contained
        in any of the choices.

    icon: bool, default True
        If `True`, include the question icon.

    warn: bool, default True
        If `True`, raise warnings when invalid input is entered.

    noask: bool, default False
        If `True`, skip printing the question and return the default value.

    Returns
    -------
    A string for a single answer or a tuple of strings if `multiple` is `True`.

    """
    from meerschaum.utils.warnings import warn as _warn
    from meerschaum.utils.packages import attempt_import
    noask = check_noask(noask)

    ### Handle empty choices.
    if len(choices) == 0:
        _warn(f"No available choices. Returning default value '{default}'.", stacklevel=3)
        return default

    ### If the default case is to include multiple answers, allow for multiple inputs.
    if isinstance(default, list):
        multiple = True

    def _enforce_default(d):
        if d is not None and d not in choices and warn:
            _warn(
                f"Default choice '{default}' is not contained in the choices {choices}. "
                + "Setting numeric = False.",
                stacklevel = 3
            )
            return False
        return True

    ### Throw a warning if the default isn't a choice.
    for d in (default if isinstance(default, list) else [default]):
        if not _enforce_default(d):
            numeric = False
            break

    _default = default
    _choices = choices
    if multiple:
        question += f"\n    Enter your choices, separated by '{delimiter}'."

    altered_choices = {}
    altered_indices = {}
    altered_default_indices = {}
    delim_replacement = '_' if delimiter != '_' else '-'
    can_strip_start_spaces, can_strip_end_spaces = True, True
    for c in choices:
        if can_strip_start_spaces and c.startswith(' '):
            can_strip_start_spaces = False
        if can_strip_end_spaces and c.endswith(' '):
            can_strip_end_spaces = False

    if multiple:
        ### Check if the defaults have the delimiter.
        for i, d in enumerate(default if isinstance(default, list) else [default]):
            if d is None or delimiter not in d:
                continue
            new_d = d.replace(delimiter, delim_replacement)
            altered_choices[new_d] = d
            altered_default_indices[i] = new_d
        for i, new_d in altered_default_indices.items():
            if not isinstance(default, list):
                default = new_d
                break
            default[i] = new_d

        ### Check if the choices have the delimiter.
        for i, c in enumerate(choices):
            if delimiter in c and warn:
                _warn(
                    f"The delimiter '{delimiter}' is contained within choice '{c}'.\n"
                    + f"Replacing the string '{delimiter}' with '{delim_replacement}' in "
                    + "the choice for correctly parsing input (will be replaced upon returning the prompt).",
                    stacklevel = 3,
                )
                new_c = c.replace(delimiter, delim_replacement)
                altered_choices[new_c] = c
                altered_indices[i] = new_c
        for i, new_c in altered_indices.items():
            choices[i] = new_c
        default = delimiter.join(default) if isinstance(default, list) else default

    if numeric:
        _choices = [str(i + 1) for i, c in enumerate(choices)]
        _default = ''
        if default is not None:
            for d in (default.split(delimiter) if multiple else [default]):
                _d = str(choices.index(d) + 1)
                _default += _d + delimiter
        _default = _default[:-1 * len(delimiter)]
        question += '\n'
        choices_digits = len(str(len(choices)))
        for i, c in enumerate(choices):
            question += f"  {i + 1}. " + (" " * (choices_digits - len(str(i + 1)))) + f"{c}\n"
        default_tuple = (_default, default) if default is not None else None
    else:
        default_tuple = default
        question += '\n'
        for c in choices:
            question += f"  - {c}\n"

    if 'completer' not in kw:
        WordCompleter = attempt_import('prompt_toolkit.completion').WordCompleter
        kw['completer'] = WordCompleter(choices, sentence=True)

    valid = False
    while not valid:
        answer = prompt(
            question,
            icon = icon,
            default = default_tuple,
            noask = noask,
            **kw
        )
        ### Split along the delimiter.
        _answers = [answer] if not multiple else [a for a in answer.split(delimiter)]

        ### Remove trailing spaces if possible.
        _answers = [(_a.rstrip(' ') if can_strip_end_spaces else _a) for _a in _answers]

        ### Remove leading spaces if possible.
        _answers = [(_a.lstrip(' ') if can_strip_start_spaces else _a) for _a in _answers]

        ### Remove empty strings.
        _answers = [_a for _a in _answers if _a]

        if multiple and len(_answers) == 0:
            _answers = default_tuple if isinstance(default_tuple, list) else [default_tuple]
        answers = [altered_choices.get(a, a) for a in _answers]

        valid = (len(answers) > 1 or not (len(answers) == 1 and answers[0] is None))
        for a in answers:
            if (
                not a in {_original for _new, _original in altered_choices.items()}
                and not a in _choices
                and a != default
                and not noask
            ):
                valid = False
                break
        if valid:
            break
        if warn:
            _warn(f"Please pick a valid choice.", stack=False)

    if not multiple:
        if not numeric:
            return answer
        try:
            return choices[int(answer) - 1]
        except Exception as e:
            _warn(f"Could not cast answer '{answer}' to an integer.", stacklevel=3)

    if not numeric:
        return answers
    _answers = []
    for a in answers:
        try:
            _answer = choices[int(a) - 1]
            _answers.append(altered_choices.get(_answer, _answer))
        except Exception as e:
            _warn(f"Could not cast answer '{a}' to an integer.", stacklevel=3)
    return _answers


def get_password(
        username: Optional[str] = None,
        minimum_length: Optional[int] = None,
        confirm: bool = True,
        **kw: Any
    ) -> str:
    """
    Prompt the user for a password.

    Parameters
    ----------
    username: Optional[str], default None
        If provided, print the username when asking for a password.

    minimum_length: Optional[int], default None
        If provided, enforce a password of at least this length.

    confirm: bool, default True
        If `True`, prompt the user for a password twice.

    Returns
    -------
    The password string (censored from terminal output when typing).

    Examples
    --------
    ```python-repl
    >>> get_password()
     ❓ Password: *******
     ❓ Confirm password: *******
    'hunter2'
    ```

    """
    from meerschaum.utils.warnings import warn
    while True:
        password = prompt(
            f"Password" + (f" for user '{username}':" if username is not None else ":"),
            is_password = True,
            **kw
        )
        if minimum_length is not None and len(password) < minimum_length:
            warn(
                "Password is too short. " +
                f"Please enter a password that is at least {minimum_length} characters.",
                stack = False
            )
            continue

        if not confirm:
            return password

        _password = prompt(
            f"Confirm password" + (f" for user '{username}':" if username is not None else ":"),
            is_password = True,
            **kw
        )
        if password != _password:
            warn(f"Passwords do not match! Please try again.", stack=False)
            continue
        else:
            return password


def get_email(username: Optional[str] = None, allow_omit: bool = True, **kw: Any) -> str:
    """
    Prompt the user for an email and enforce that it's valid.

    Parameters
    ----------
    username: Optional[str], default None
        If provided, print the username in the prompt.

    allow_omit: bool, default True
        If `True`, allow the user to omit the email.

    Returns
    -------
    The provided email string.

    Examples
    --------
    ```python-repl
    >>> get_email()
     ❓ Email (empty to omit): foo@foo
     Invalid email! Please try again.
     ❓ Email (empty to omit): foo@foo.com
    'foo@foo.com'
    ```
    """
    from meerschaum.utils.warnings import warn
    from meerschaum.utils.misc import is_valid_email
    while True:
        email = prompt(
            f"Email" + (f" for user '{username}'" if username is not None else "") +
            (" (empty to omit):" if allow_omit else ": "),
            **kw
        )
        if (allow_omit and email == '') or is_valid_email(email):
            return email
        warn(f"Invalid email! Please try again.", stack=False)


def check_noask(noask: bool = False) -> bool:
    """
    Flip `noask` to `True` if `MRSM_NOASK` is set.
    """
    from meerschaum.config.static import STATIC_CONFIG
    NOASK = STATIC_CONFIG['environment']['noask']
    if noask:
        return True
    return (
        os.environ.get(NOASK, 'false').lower()
        in ('1', 'true')
    )



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Handle all things warnings and errors here
"""

from __future__ import annotations
from meerschaum.utils.typing import Any, Union

import sys
import warnings

warnings.filterwarnings(
    "always",
    category = UserWarning
)
warnings.filterwarnings(
    "always",
    category = ImportWarning
)
warnings.filterwarnings(
    "ignore",
    category = RuntimeWarning
)


def enable_depreciation_warnings(name) -> None:
    """Enable depreciation warnings in the warnings module."""
    import meerschaum.actions
    warnings.filterwarnings(
        "always",
        category = DeprecationWarning,
        module = name
    )

def warn(*args, stacklevel=2, stack=True, color: bool = True, **kw) -> None:
    """
    Print a stylized warning message.
    May be captured by `warnings.filterwarnings()`.
    """
    if stacklevel is None:
        stacklevel = 1
        stack = False
    _old_sw = warnings.showwarning

    get_config = None
    if color:
        try:
            from meerschaum.utils.formatting import (
                CHARSET, ANSI, colored, fill_ansi, highlight_pipes, _init
            )
        except ImportError:
            CHARSET = 'ascii'
            ANSI = False
        try:
            from meerschaum.config import get_config as _get_config
            from meerschaum.config import _config
            cf = _config()
            get_config = _get_config
        except ImportError:
            get_config = None

    if get_config is None and color:
        try:
            warn_config = cf['formatting']['warnings']
        except KeyError:
            warn_config = {
                'ansi' : {'color' : []},
                'unicode' : {'icon' : ''},
                'ascii' : {'icon' : ''},
            }
    elif color:
        warn_config = get_config('formatting', 'warnings', patch=True)
    a = list(args)
    a[0] = ' ' + (warn_config[CHARSET]['icon'] if color else '') + ' ' + str(a[0])
    if color:
        if ANSI:
            _init()
            a[0] = fill_ansi(highlight_pipes(a[0]), **warn_config['ansi']['rich'])

    ### Optionally omit the warning location.
    def _no_stack_sw(message, category, filename, lineno, file=None, line=None):
        sys.stderr.write(str(message) + '\n')

    if not stack:
        warnings.showwarning = _no_stack_sw
    warnings.warn(*a, stacklevel=stacklevel, **kw)
    if not stack:
        warnings.showwarning = _old_sw


def exception_with_traceback(
        message: str,
        exception_class = Exception, 
        stacklevel = 1,
        tb_type = 'single'
    ):
    """Traceback construction help found here:
    https://stackoverflow.com/questions/27138440/how-to-create-a-traceback-object
    """
    import types
    tb, depth = None, 0
    while True:
        try:
            frame = sys._getframe(depth)
            depth += 1
        except ValueError as e:
            break

        tb = types.TracebackType(tb, frame, frame.f_lasti, frame.f_lineno)

    tbs, _tb = [], tb
    while True:
        if _tb is None:
            break
        tbs.append(_tb)
        _tb = _tb.tb_next

    found_main, main_i = False, 0
    first_mrsm_after_main = None
    last_mrsm_i = None
    tbs[(-1 * stacklevel)].tb_next = None
    for i, _tb in enumerate([_tb for _tb in tbs]):
        if 'meerschaum' in str(_tb.tb_frame) and '__main__.py' in str(_tb.tb_frame):
            found_main = True
            main_i = i
            continue
        if i >= (len(tbs) - (stacklevel - 1)):
            tbs[i] = None
        elif (
                found_main and 'meerschaum' in str(_tb.tb_frame)
                and first_mrsm_after_main is None
                and 'Shell' not in str(_tb.tb_frame)
            ):
            first_mrsm_after_main = i
        elif 'meerschaum' in str(_tb.tb_frame):
            last_mrsm_i = i

    tbs = [_tb for _tb in tbs if tb is not None]

    if tb_type == 'single':
        return exception_class(message).with_traceback(tbs[-3])
    return exception_class(message).with_traceback(tbs[first_mrsm_after_main])


def error(
        message: str,
        exception_class = Exception,
        nopretty: bool = False,
        silent: bool = True,
        stack: bool = True,
        raise_: bool = True,
    ):
    """
    Raise an exception with supressed traceback.
    """
    from meerschaum.utils.formatting import (
        CHARSET, ANSI, get_console, fill_ansi, highlight_pipes, _init
    )
    from meerschaum.utils.packages import import_rich, attempt_import
    from meerschaum.config import get_config
    import types, inspect
    rich = import_rich()
    rich_traceback = attempt_import('rich.traceback')
    error_config = get_config('formatting', 'errors', patch=True)
    message = ' ' + error_config[CHARSET]['icon'] + ' ' + str(message)
    exception = exception_with_traceback(message, exception_class, stacklevel=3)
    color_message = str(message)
    color_exception = exception_with_traceback(color_message, exception_class, stacklevel=3)
    if ANSI and not nopretty:
        _init()
        color_message = '\n' + fill_ansi(highlight_pipes(message), **error_config['ansi']['rich'])
        color_exception = exception_with_traceback(color_message, exception_class, stacklevel=3)
    try:
        trace = rich_traceback.Traceback.extract(
            exception_class, exception, exception.__traceback__
        )
        rtb = rich_traceback.Traceback(trace)
    except Exception as e:
        trace, rtb = None, None
    if trace is None or rtb is None:
        nopretty = True
    if not nopretty and stack:
        if get_console() is not None:
            get_console().print(rtb)
    frame = sys._getframe(len(inspect.stack()) - 1)
    if raise_:
        raise color_exception


def info(message: str, icon: bool = True, **kw):
    """Print an informative message."""
    from meerschaum.utils.packages import import_rich, attempt_import
    from meerschaum.utils.formatting import (
        CHARSET, ANSI, highlight_pipes, fill_ansi, _init
    )
    from meerschaum.config import get_config
    info_config = get_config('formatting', 'info', patch=True)
    if icon:
        message = ' ' + info_config[CHARSET]['icon'] + ' ' + message
    if ANSI:
        _init()
        message = highlight_pipes(message)
        lines = message.split('\n')
        message = fill_ansi(lines[0], **info_config['ansi']['rich']) + (
            '\n' + '\n'.join(lines[1:]) if len(lines) > 1 else ''
        )
    print(message)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Implement the Connector fetch() method
"""

from __future__ import annotations
from datetime import datetime, timedelta
import meerschaum as mrsm
from meerschaum.utils.typing import Optional, Union, Callable, Any

def fetch(
        self,
        pipe: meerschaum.Pipe,
        begin: Union[datetime, int, str, None] = '',
        end: Union[datetime, int, str, None] = None,
        check_existing: bool = True,
        chunk_hook: Optional[Callable[[pd.DataFrame], Any]] = None,
        chunksize: Optional[int] = -1,
        workers: Optional[int] = None,
        debug: bool = False,
        **kw: Any
    ) -> Union['pd.DataFrame', List[Any], None]:
    """Execute the SQL definition and return a Pandas DataFrame.

    Parameters
    ----------
    pipe: meerschaum.Pipe
        The pipe object which contains the `fetch` metadata.
        
        - pipe.columns['datetime']: str
            - Name of the datetime column for the remote table.
        - pipe.parameters['fetch']: Dict[str, Any]
            - Parameters necessary to execute a query.
        - pipe.parameters['fetch']['definition']: str
            - Raw SQL query to execute to generate the pandas DataFrame.
        - pipe.parameters['fetch']['backtrack_minutes']: Union[int, float]
            - How many minutes before `begin` to search for data (*optional*).

    begin: Union[datetime, int, str, None], default None
        Most recent datatime to search for data.
        If `backtrack_minutes` is provided, subtract `backtrack_minutes`.

    end: Union[datetime, int, str, None], default None
        The latest datetime to search for data.
        If `end` is `None`, do not bound 

    check_existing: bool, defult True
        If `False`, use a backtrack interval of 0 minutes.

    chunk_hook: Callable[[pd.DataFrame], Any], default None
        A function to pass to `SQLConnector.read()` that accepts a Pandas DataFrame.

    chunksize: Optional[int], default -1
        How many rows to load into memory at once (when `chunk_hook` is provided).
        Otherwise the entire result set is loaded into memory.

    workers: Optional[int], default None
        How many threads to use when consuming the generator (when `chunk_hook is provided).
        Defaults to the number of cores.

    debug: bool, default False
        Verbosity toggle.
       
    Returns
    -------
    A pandas DataFrame or `None`.
    If `chunk_hook` is not None, return a list of the hook function's results.
    """
    meta_def = self.get_pipe_metadef(
        pipe,
        begin = begin,
        end = end,
        check_existing = check_existing,
        debug = debug,
        **kw
    )
    as_hook_results = chunk_hook is not None
    chunks = self.read(
        meta_def,
        chunk_hook = chunk_hook,
        as_hook_results = as_hook_results,
        chunksize = chunksize,
        workers = workers,
        debug = debug,
    )
    ### if sqlite, parse for datetimes
    if not as_hook_results and self.flavor == 'sqlite':
        from meerschaum.utils.misc import parse_df_datetimes
        ignore_cols = [
            col
            for col, dtype in pipe.dtypes.items()
            if 'datetime' not in str(dtype)
        ]
        return (
            parse_df_datetimes(
                chunk,
                ignore_cols = ignore_cols,
                debug = debug,
            )
            for chunk in chunks
        )
    return chunks


def get_pipe_metadef(
        self,
        pipe: meerschaum.Pipe,
        params: Optional[Dict[str, Any]] = None,
        begin: Union[datetime, int, str, None] = '',
        end: Union[datetime, int, str, None] = None,
        check_existing: bool = True,
        debug: bool = False,
        **kw: Any
    ) -> Union[str, None]:
    """
    Return a pipe's meta definition fetch query.

    params: Optional[Dict[str, Any]], default None
        Optional params dictionary to build the `WHERE` clause.
        See `meerschaum.utils.sql.build_where`.

    begin: Union[datetime, int, str, None], default None
        Most recent datatime to search for data.
        If `backtrack_minutes` is provided, subtract `backtrack_minutes`.

    end: Union[datetime, int, str, None], default None
        The latest datetime to search for data.
        If `end` is `None`, do not bound 

    check_existing: bool, default True
        If `True`, apply the backtrack interval.

    debug: bool, default False
        Verbosity toggle.

    Returns
    -------
    A pipe's meta definition fetch query string.
    """
    from meerschaum.utils.debug import dprint
    from meerschaum.utils.warnings import warn, error
    from meerschaum.utils.sql import sql_item_name, dateadd_str, build_where
    from meerschaum.utils.misc import is_int
    from meerschaum.config import get_config

    definition = get_pipe_query(pipe)

    if not pipe.columns.get('datetime', None):
        _dt = pipe.guess_datetime()
        dt_name = sql_item_name(_dt, self.flavor, None) if _dt else None
        is_guess = True
    else:
        _dt = pipe.get_columns('datetime')
        dt_name = sql_item_name(_dt, self.flavor, None)
        is_guess = False

    if begin not in (None, '') or end is not None:
        if is_guess:
            if _dt is None:
                warn(
                    f"Unable to determine a datetime column for {pipe}."
                    + "\n    Ignoring begin and end...",
                    stack = False,
                )
                begin, end = '', None
            else:
                warn(
                    f"A datetime wasn't specified for {pipe}.\n"
                    + f"    Using column \"{_dt}\" for datetime bounds...",
                    stack = False
                )


    apply_backtrack = begin == '' and check_existing
    backtrack_interval = pipe.get_backtrack_interval(check_existing=check_existing, debug=debug)
    btm = (
        int(backtrack_interval.total_seconds() / 60)
        if isinstance(backtrack_interval, timedelta)
        else backtrack_interval
    )
    begin = (
        pipe.get_sync_time(debug=debug)
        if begin == ''
        else begin
    )

    if begin and end and begin >= end:
        begin = None
    
    da = None
    if dt_name:
        begin_da = (
            dateadd_str(
                flavor = self.flavor,
                datepart = 'minute',
                number = ((-1 * btm) if apply_backtrack else 0), 
                begin = begin,
            )
            if begin
            else None
        )
        end_da = (
            dateadd_str(
                flavor = self.flavor,
                datepart = 'minute',
                number = 0,
                begin = end,
            )
            if end
            else None
        )

    meta_def = (
        _simple_fetch_query(pipe) if (
            (not (pipe.columns or {}).get('id', None))
            or (not get_config('system', 'experimental', 'join_fetch'))
        ) else _join_fetch_query(pipe, debug=debug, **kw)
    )

    has_where = 'where' in meta_def.lower()[meta_def.lower().rfind('definition'):]
    if dt_name and (begin_da or end_da):
        definition_dt_name = (
            dateadd_str(self.flavor, 'minute', 0, f"definition.{dt_name}")
            if not is_int((begin_da or end_da))
            else f"definition.{dt_name}"
        )
        meta_def += "\n" + ("AND" if has_where else "WHERE") + " "
        has_where = True
        if begin_da:
            meta_def += f"{definition_dt_name} >= {begin_da}"
        if begin_da and end_da:
            meta_def += " AND "
        if end_da:
            meta_def += f"{definition_dt_name} < {end_da}"

    if params is not None:
        params_where = build_where(params, self, with_where=False)
        meta_def += "\n" + ("AND" if has_where else "WHERE") + " "
        has_where = True
        meta_def += params_where

    return meta_def


def get_pipe_query(pipe: mrsm.Pipe, warn: bool = True) -> Union[str, None]:
    """
    Run through the possible keys for a pipe's query and return the first match.

    - fetch, definition
    - definition
    - query
    - sql
    """
    from meerschaum.utils.warnings import warn as _warn
    if pipe.parameters.get('fetch', {}).get('definition', None):
        definition = pipe.parameters['fetch']['definition']
    elif pipe.parameters.get('definition', None):
        definition = pipe.parameters['definition']
    elif pipe.parameters.get('query', None):
        definition = pipe.parameters['query']
    elif pipe.parameters.get('sql', None):
        definition = pipe.parameters['sql']
    else:
        if warn:
            _warn(
                f"Could not determine a SQL definition for {pipe}.\n"
                + "    Set the key `query` in `pipe.parameters` to a valid SQL query."
            )
        return None
    return definition


def set_pipe_query(pipe: mrsm.Pipe, query: str) -> None:
    """
    Run through the possible keys for a pipe's query and set the first match.

    - fetch, definition
    - definition
    - query
    - sql
    """
    if pipe.parameters.get('fetch', {}).get('definition', None):
        if pipe.parameters.get('fetch', None) is None:
            pipe.parameters['fetch'] = {}
        dict_to_set = pipe.parameters['fetch']
        key_to_set = 'definition'
    elif pipe.parameters.get('definition', None):
        dict_to_set = pipe.parameters
        key_to_set = 'definition'
    elif pipe.parameters.get('query', None):
        dict_to_set = pipe.parameters
        key_to_set = 'query'
    else:
        dict_to_set = pipe.parameters
        key_to_set = 'sql'

    dict_to_set[key_to_set] = query


def _simple_fetch_query(pipe, debug: bool=False, **kw) -> str:
    """Build a fetch query from a pipe's definition."""
    def_name = 'definition'
    definition = get_pipe_query(pipe)
    return (
        f"WITH {def_name} AS (\n{definition}\n) SELECT * FROM {def_name}"
        if pipe.connector.flavor not in ('mysql', 'mariadb')
        else f"SELECT * FROM (\n{definition}\n) AS {def_name}"
    )

def _join_fetch_query(
        pipe,
        debug: bool = False,
        new_ids: bool = True,
        **kw
    ) -> str:
    """Build a fetch query based on the datetime and ID indices."""
    if not pipe.exists(debug=debug):
        return _simple_fetch_query(pipe, debug=debug, **kw)

    from meerschaum.utils.sql import sql_item_name, dateadd_str
    pipe_instance_name = sql_item_name(
        pipe.target, pipe.instance_connector.flavor, pipe.instance_connector.schema
    )
    #  pipe_remote_name = sql_item_name(pipe.target, pipe.connector.flavor)
    sync_times_table = pipe.target + "_sync_times"
    sync_times_instance_name = sql_item_name(
        sync_times_table, pipe.instance_connector.flavor, None
    )
    sync_times_remote_name = sql_item_name(
        sync_times_table, pipe.connector.flavor, None
    )
    id_instance_name = sql_item_name(
        pipe.columns['id'], pipe.instance_connector.flavor, None
    )
    id_remote_name = sql_item_name(pipe.columns['id'], pipe.connector.flavor, None)
    dt_instance_name = sql_item_name(
        pipe.columns['datetime'], pipe.instance_connector.flavor, None
    )
    dt_remote_name = sql_item_name(
        pipe.columns['datetime'], pipe.connector.flavor, None
    )
    cols_types = pipe.get_columns_types(debug=debug)
    sync_times_query = f"""
    SELECT {id_instance_name}, MAX({dt_instance_name}) AS {dt_instance_name}
    FROM {pipe_instance_name}
    GROUP BY {id_instance_name}
    """
    sync_times = pipe.instance_connector.read(sync_times_query, debug=debug, silent=False)
    if sync_times is None:
        return _simple_fetch_query(pipe, debug=debug, **kw)
    _sync_times_q = f",\n{sync_times_remote_name} AS ("
    for _id, _st in sync_times.itertuples(index=False):
        _sync_times_q += (
            f"SELECT CAST('{_id}' AS "
            + sql_item_name(
                cols_types[pipe.columns['id']], pipe.connector.flavor, None
            )
            + f") AS {id_remote_name}, "
            + dateadd_str(
                flavor = pipe.connector.flavor,
                begin = _st,
                datepart = 'minute',
                number = pipe.parameters.get('fetch', {}).get('backtrack_minutes', 0)
            ) + " AS " + dt_remote_name + "\nUNION ALL\n"
        )
    _sync_times_q = _sync_times_q[:(-1 * len('UNION ALL\n'))] + ")"

    definition = get_pipe_query(pipe)
    query = (
        f"""
    WITH definition AS (\n{definition}\n){_sync_times_q}
    SELECT definition.*
    FROM definition"""
        if pipe.connector.flavor not in ('mysql', 'mariadb')
        else (
        f"""
    SELECT * FROM (\n{definition}\n) AS definition"""
        )
    ) + f"""
    LEFT OUTER JOIN {sync_times_remote_name} AS st
      ON st.{id_remote_name} = definition.{id_remote_name}
    WHERE definition.{dt_remote_name} > st.{dt_remote_name}
    """ + (f"  OR st.{id_remote_name} IS NULL" if new_ids else "")
    return query




#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Directory of necessary packages

packages dictionary is structured in the following schema:
    {
        <group> : {
            <import_name> : <install_name>
        }
    }
"""

from __future__ import annotations
from meerschaum.utils.typing import Dict

packages: Dict[str, Dict[str, str]] = {
    'required': {
    },
    'minimal': {},
    'formatting': {
        'pprintpp'                   : 'pprintpp>=0.4.0',
        'asciitree'                  : 'asciitree>=0.3.3',
        'typing_extensions'          : 'typing-extensions>=4.7.1',
        'pygments'                   : 'pygments>=2.7.2',
        'colorama'                   : 'colorama>=0.4.3',
        'rich'                       : 'rich>=13.4.2',
        'more_termcolor'             : 'more-termcolor>=1.1.3',
        'humanfriendly'              : 'humanfriendly>=10.0.0',
    },
    '_required': {
        'wheel'                      : 'wheel>=0.34.2',
        'setuptools'                 : 'setuptools>=63.3.0',
        'yaml'                       : 'PyYAML>=5.3.1',
        'pip'                        : 'pip>=22.0.4',
        'update_checker'             : 'update-checker>=0.18.0',
        'semver'                     : 'semver>=3.0.0',
        'pathspec'                   : 'pathspec>=0.9.0',
        'dateutil'                   : 'python-dateutil>=2.7.5',
        'requests'                   : 'requests>=2.23.0',
        'binaryornot'                : 'binaryornot>=0.4.4',
        'pyvim'                      : 'pyvim>=3.0.2',
        'aiofiles'                   : 'aiofiles>=0.6.0',
        'packaging'                  : 'packaging>=21.3.0',
        'prompt_toolkit'             : 'prompt-toolkit>=3.0.39',
        'more_itertools'             : 'more-itertools>=8.7.0',
        'daemon'                     : 'python-daemon>=0.2.3',
        'fasteners'                  : 'fasteners>=0.18.0',
        'psutil'                     : 'psutil>=5.8.0',
        'watchgod'                   : 'watchgod>=0.7.0',
        'dill'                       : 'dill>=0.3.3',
        'virtualenv'                 : 'virtualenv>=20.1.0',
        'rocketry'                   : 'rocketry>=2.5.1',
    },
    'drivers': {
        'cryptography'               : 'cryptography>=38.0.1',
        'psycopg'                    : 'psycopg[binary]>=3.1.18',
        'pymysql'                    : 'PyMySQL>=0.9.0',
        'aiomysql'                   : 'aiomysql>=0.0.21',
        'sqlalchemy_cockroachdb'     : 'sqlalchemy-cockroachdb>=2.0.0',
        'duckdb'                     : 'duckdb>=0.9.0',
        'duckdb_engine'              : 'duckdb-engine>=0.9.2',
    },
    '_drivers': {
        'pyodbc'                     : 'pyodbc>=4.0.30',
        'cx_Oracle'                  : 'cx_Oracle>=8.3.0',
    },
    'cli': {
        'pgcli'                      : 'pgcli>=3.1.0',
        'mycli'                      : 'mycli>=1.23.2',
        'litecli'                    : 'litecli>=1.5.0',
        'mssqlcli'                   : 'mssql-cli>=1.0.0',
        'gadwall'                    : 'gadwall>=0.2.0',
    },
    'stack': {
        'compose'                    : 'docker-compose>=1.27.4',
    },
    'build': {
        'cx_Freeze'                  : 'cx_Freeze>=6.5.1',
        'PyInstaller'                : 'pyinstaller>=5.0.0-dev0',
    },
    'dev-tools': {
        'twine'                      : 'twine>=3.2.0',
        'tuna'                       : 'tuna>=0.5.3',
        'snakeviz'                   : 'snakeviz>=2.1.0',
        'mypy'                       : 'mypy>=0.812.0',
        'pytest'                     : 'pytest>=6.2.2',
        'pytest_xdist'               : 'pytest-xdist>=3.2.1',
        'heartrate'                  : 'heartrate>=0.2.1',
    },
    'setup': {
    },
    'docs': {
        'mkdocs'                     : 'mkdocs>=1.1.2',
        'mkdocs_material'            : 'mkdocs-material>=6.2.5',
        'mkdocs_material_extensions' : 'mkdocs-material-extensions>=1.0.3',
        'mkdocs_autolinks_plugin'    : 'mkdocs-autolinks-plugin>=0.2.0',
        'mkdocs_awesome_pages_plugin': 'mkdocs-awesome-pages-plugin>=2.5.0',
        'mkdocs_section_index'       : 'mkdocs-section-index>=0.3.3',
        'mkdocs_linkcheck'           : 'mkdocs-linkcheck>=1.0.6',
        'mkdocs_redirects'           : 'mkdocs-redirects>=1.0.4',
        'jinja2'                     : 'jinja2==3.0.3',
    },
    'gui': {
        'toga'                       : 'toga>=0.3.0-dev29',
        'webview'                    : 'pywebview>=3.6.3',
        'pycparser'                  : 'pycparser>=2.21.0',
    },
    'extras': {
        'cmd2'                       : 'cmd2>=1.4.0',
        'ruamel.yaml'                : 'ruamel.yaml>=0.16.12',
        'modin'                      : 'modin[ray]>=0.8.3',
        'nanoid'                     : 'nanoid>=2.0.0',
        'importlib_metadata'         : 'importlib-metadata>=4.12.0',
    },
}
packages['sql'] = {
    'numpy'                          : 'numpy>=1.18.5',
    'pandas'                         : 'pandas[parquet]>=2.0.1',
    'pyarrow'                        : 'pyarrow>=7.0.0',
    'dask'                           : 'dask>=2023.5.0',
    'pytz'                           : 'pytz',
    'joblib'                         : 'joblib>=0.17.0',
    'sqlalchemy'                     : 'SQLAlchemy>=2.0.5',
    'databases'                      : 'databases>=0.4.0',
    'aiosqlite'                      : 'aiosqlite>=0.16.0',
    'asyncpg'                        : 'asyncpg>=0.21.0',
}
packages['sql'].update(packages['drivers'])
packages['sql'].update(packages['_required'])
packages['dash'] = {
    'flask_compress'                 : 'Flask-Compress>=1.10.1',
    'dash'                           : 'dash>=2.6.2',
    'dash_bootstrap_components'      : 'dash-bootstrap-components>=1.2.1',
    'dash_ace'                       : 'dash-ace>=0.2.1',
    'dash_extensions'                : 'dash-extensions>=1.0.4',
    'dash_daq'                       : 'dash-daq>=0.5.0',
    'terminado'                      : 'terminado>=0.12.1',
    'tornado'                        : 'tornado>=6.1.0',
}
packages['api'] = {
    'uvicorn'                        : 'uvicorn[standard]>=0.22.0',
    'gunicorn'                       : 'gunicorn>=20.1.0',
    'dotenv'                         : 'python-dotenv>=0.20.0',
    'websockets'                     : 'websockets>=11.0.3',
    'fastapi'                        : 'fastapi>=0.100.0',
    'passlib'                        : 'passlib>=1.7.4',
    'fastapi_login'                  : 'fastapi-login>=1.7.2',
    'multipart'                      : 'python-multipart>=0.0.5',
    'pydantic'                       : 'pydantic<2.0.0',
    'httpx'                          : 'httpx>=0.24.1',
    'websockets'                     : 'websockets>=11.0.3',
}
packages['api'].update(packages['sql'])
packages['api'].update(packages['formatting'])
packages['api'].update(packages['dash'])

all_packages = {}
for group, import_names in packages.items():
    all_packages.update(import_names)
install_names = {}
def get_install_names():
    """
    Initialize the mapping between install names and import names.
    """
    if not install_names:
        from meerschaum.utils.packages import get_install_no_version
        for _import_name, _install_name in all_packages.items():
            install_names[get_install_no_version(_install_name)] = _import_name
    return install_names

skip_groups = {'docs', 'build', 'cli', 'dev-tools', 'portable', 'extras', 'stack', '_drivers'}
full = []
_full = {}
for group, import_names in packages.items():
    ### omit 'cli' and 'docs' from 'full'
    if group in skip_groups:
        continue
    full += [ install_name for import_name, install_name in import_names.items() ]
    for import_name, install_name in import_names.items():
        _full[import_name] = install_name
packages['full'] = _full




#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Define callbacks for pages.
"""

import meerschaum.api.dash.callbacks.dashboard
import meerschaum.api.dash.callbacks.login
import meerschaum.api.dash.callbacks.plugins
import meerschaum.api.dash.callbacks.jobs
import meerschaum.api.dash.callbacks.register



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Import the core class definitions into the parent module.
"""

from meerschaum.core.Pipe import Pipe
from meerschaum.core.Plugin import Plugin
from meerschaum.core.User import User



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Default actions available to the mrsm CLI.
"""

from __future__ import annotations
from meerschaum.utils.typing import Callable, Any, Optional, Union, List, Dict, SuccessTuple
from meerschaum.utils.packages import get_modules_from_package
_custom_actions = []

def get_subactions(
        action: Union[str, List[str]],
        _actions: Optional[Dict[str, Callable[[Any], Any]]] = None,
    ) -> Dict[str, Callable[[Any], Any]]:
    """
    Return a dictionary of an action's sub-action functions.

    Examples
    --------
    >>> get_subactions('install').keys()
    dict_keys(['packages', 'plugins', 'required'])
    >>> 
    """
    import importlib, inspect
    subactions = {}
    if isinstance(action, str):
        action = [action]
    action_function = get_action(action[0], _actions=_actions)
    if action_function is None:
        return subactions
    try:
        action_module = importlib.import_module(action_function.__module__)
    except ImportError:
        action_module = None
    if action_module is None:
        return subactions
    for name, f in inspect.getmembers(action_module):
        if not inspect.isfunction(f):
            continue
        if action_function.__name__ + '_' in name and not name.lstrip('_').startswith('complete'):
            _name = name.replace(action_function.__name__, '')
            _name = _name.lstrip('_')
            subactions[_name] = f
    return subactions


def get_action(
        action: Union[str, List[str]],
        _actions: Optional[Dict[str, Callable[[Any], Any]]] = None,
    ) -> Union[Callable[[Any], Any], None]:
    """
    Return a function corresponding to the given action list.
    This may be a custom action with an underscore, in which case, allow for underscores.
    This may also be a subactions, which is handled by `get_subactions()`
    """
    if _actions is None:
        _actions = actions
    if isinstance(action, str):
        action = [action]

    if not any(action):
        return None

    ### Simple case, e.g. ['show']
    if len(action) == 1:
        if action[0] in _actions:
            return _actions[action[0]]

        ### e.g. ['foo'] (and no custom action available)
        return None

    ### Last case: it could be a custom action with an underscore in the name.
    action_name_with_underscores = '_'.join(action)
    candidates = []
    for action_key, action_function in _actions.items():
        if not '_' in action_key:
            continue
        if action_name_with_underscores.startswith(action_key):
            leftovers = action_name_with_underscores.replace(action_key, '')
            candidates.append((len(leftovers), action_function))
    if len(candidates) > 0:
        return sorted(candidates)[0][1]

    ### Might be dealing with a subaction.
    if action[0] in _actions:
        subactions = get_subactions([action[0]], _actions=_actions)
        if action[1] not in subactions:
            return _actions[action[0]]
        return subactions[action[1]]

    return None


def get_main_action_name(
        action: Union[List[str], str],
        _actions: Optional[Dict[str, Callable[[Any], SuccessTuple]]] = None,
    ) -> Union[str, None]:
    """
    Given an action list, return the name of the main function.
    For subactions, this will return the root function.
    If no action function can be found, return `None`.

    Parameters
    ----------
    action: Union[List[str], str]
        The standard action list.

    Examples
    --------
    >>> get_main_action_name(['show', 'pipes'])
    'show'
    >>> 
    """
    if _actions is None:
        _actions = actions
    action_function = get_action(action, _actions=_actions)
    if action_function is None:
        return None
    clean_name = action_function.__name__.lstrip('_')
    if clean_name in _actions:
        return clean_name
    words = clean_name.split('_')
    first_word = words[0]
    if first_word in _actions:
        return first_word
    ### We might be dealing with shell `do_` functions.
    second_word = words[1] if len(words) > 1 else None
    if second_word and second_word in _actions:
        return second_word
    return None


def get_completer(
        action: Union[List[str], str],
        _actions: Optional[Dict[str, Callable[[Any], SuccessTuple]]] = None,
    ) -> Union[
        Callable[['meerschaum._internal.shell.Shell', str, str, int, int], List[str]], None
    ]:
    """Search for a custom completer function for an action."""
    import importlib, inspect
    if isinstance(action, str):
        action = [action]
    action_function = get_action(action)
    if action_function is None:
        return None
    try:
        action_module = importlib.import_module(action_function.__module__)
    except ImportError:
        action_module = None
    if action_module is None:
        return None
    candidates = []
    for name, f in inspect.getmembers(action_module):
        if not inspect.isfunction(f):
            continue
        name_lstrip = name.lstrip('_')
        if not name_lstrip.startswith('complete_'):
            continue
        if name_lstrip == 'complete_' + action_function.__name__:
            candidates.append((name_lstrip, f))
            continue
        if name_lstrip == 'complete_' + action_function.__name__.split('_')[0]:
            candidates.append((name_lstrip, f))
    if len(candidates) == 1:
        return candidates[0][1]
    if len(candidates) > 1:
        return sorted(candidates)[-1][1]
    
    return None


def choose_subaction(
        action: Optional[List[str]] = None,
        options: Optional[Dict[str, Any]] = None,
        **kw
    ) -> SuccessTuple:
    """
    Given a dictionary of options and the standard Meerschaum actions list,
    check if choice is valid and execute chosen function, else show available
    options and return False

    Parameters
    ----------
    action: Optional[List[str]], default None
        A list of subactions (e.g. `show pipes` -> ['pipes']).

    options: Optional[Dict[str, Any]], default None
        Available options to execute.
        option (key) -> function (value)
        Functions must accept **kw keyword arguments
        and return a tuple of success bool and message.

    Returns
    -------
    The return value of the chosen subaction (assumed to be a `SuccessTuple`).

    """
    from meerschaum.utils.warnings import warn, info
    import inspect
    if action is None:
        action = []
    if options is None:
        options = {}
    parent_action = inspect.stack()[1][3]
    if len(action) == 0:
        action = ['']
    choice = action[0]

    def valid_choice(_choice : str, _options : dict):
        if _choice in _options:
            return _choice
        if (_choice + 's') in options:
            return _choice + 's'
        return None

    parsed_choice = valid_choice(choice, options)
    if parsed_choice is None:
        warn(f"Cannot {parent_action} '{choice}'. Choose one:", stack=False)
        for option in sorted(options):
            print(f"  - {parent_action} {option}")
        return (False, f"Invalid choice '{choice}'.")
    ### remove parent sub-action
    kw['action'] = list(action)
    del kw['action'][0]
    return options[parsed_choice](**kw)


def _get_subaction_names(action: str, globs: dict = None) -> List[str]:
    """Use `meerschaum.actions.get_subactions()` instead."""
    if globs is None:
        import importlib
        module = importlib.import_module(f'meerschaum.actions.{action}')
        globs = vars(module)
    subactions = []
    for item in globs:
        if f'_{action}' in item and 'complete' not in item.lstrip('_'):
            subactions.append(globs[item])
    return subactions


def choices_docstring(action: str, globs : Optional[Dict[str, Any]] = None) -> str:
    """
    Append the an action's available options to the module docstring.
    This function is to be placed at the bottom of each action module.

    Parameters
    ----------
    action: str
        The name of the action module (e.g. 'install').
        
    globs: Optional[Dict[str, Any]], default None
        An optional dictionary of global variables.

    Returns
    -------
    The generated docstring for the module.

    Examples
    --------
    >>> from meerschaum.utils.misc import choices_docstring as _choices_docstring
    >>> install.__doc__ += _choices_docstring('install')

    """
    options_str = f"\n    Options:\n        `{action} "
    subactions = _get_subaction_names(action, globs=globs)
    options_str += "["
    sa_names = []
    for sa in subactions:
        try:
            sa_names.append(sa.__name__[len(f"_{action}") + 1:])
        except Exception as e:
            print(e)
            return ""
    for sa_name in sorted(sa_names):
        options_str += f"{sa_name}, "
    options_str = options_str[:-2] + "]`"
    return options_str

### build __all__ from other .py files in this package
import sys
modules = get_modules_from_package(
    sys.modules[__name__],
    names = False,
)
__all__ = ['actions', 'get_subactions', 'get_action', 'get_main_action_name', 'get_completer']

### Build the actions dictionary by importing all
### functions that do not begin with '_' from all submodules.
from inspect import getmembers, isfunction
actions = {}

for module in modules:
    ### A couple important things happening here:
    ### 1. Find all functions in all modules in `actions` package
    ###     (skip functions that begin with '_')
    ### 2. Add them as members to the Shell class
    ###     - Original definition : meerschaum._internal.shell.Shell
    ###     - New definition      : meerschaum._internal.Shell
    ### 3. Populate the actions dictionary with function names and functions
    ###
    ### UPDATE:
    ### Shell modifications have been deferred to get_shell in order to improve lazy loading.

    actions.update(
        dict(
            [
                ### __name__ and new function pointer
                (ob[0], ob[1])
                    for ob in getmembers(module)
                        if isfunction(ob[1])
                            ### check that the function belongs to the module
                            and ob[0] == module.__name__.replace('_', '').split('.')[-1]
                            ### skip functions that start with '_'
                            and ob[0][0] != '_'
            ]
        )
    )

original_actions = actions.copy()
from meerschaum._internal.entry import entry, get_shell
import meerschaum.plugins
make_action = meerschaum.plugins.make_action
pre_sync_hook = meerschaum.plugins.pre_sync_hook
post_sync_hook = meerschaum.plugins.post_sync_hook

### Instruct pdoc to skip the `meerschaum.actions.plugins` subdirectory.
__pdoc__ = {
    'plugins': False,
    'arguments': True,
    'shell': False,
    'actions': (
        "Access functions of the standard Meerschaum actions.\n\n"
        + "Visit the [actions reference page](https://meerschaum.io/reference/actions/) "
        + "for documentation about each action.\n\n"
        + """\n\nExamples
--------
>>> actions['show'](['pipes'])

"""
        + "Keys\n----\n"
        + ', '.join(sorted([f'`{a}`' for a in actions]))
    )
}
for a in actions:
    __pdoc__[a] = False
meerschaum.plugins.load_plugins()



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Functions for interacting with the Webterm via the dashboard.
"""

import time
from urllib.parse import urlparse
from meerschaum.config import get_config
from meerschaum.api import debug, CHECK_UPDATE, get_api_connector, no_auth
from meerschaum.api.dash import active_sessions
from meerschaum.api.dash.users import is_session_authenticated
from meerschaum.api.dash.components import alert_from_success_tuple, console_div
from meerschaum.utils.typing import WebState, SuccessTuple, List, Tuple, Optional, Any
from meerschaum.utils.packages import attempt_import, import_html, import_dcc, run_python_package
from meerschaum._internal.term.tools import is_webterm_running
from meerschaum.config.static import STATIC_CONFIG
from meerschaum.utils.threading import Thread, RLock
dcc, html = import_dcc(check_update=CHECK_UPDATE), import_html(check_update=CHECK_UPDATE)
dbc = attempt_import('dash_bootstrap_components', lazy=False, check_update=CHECK_UPDATE)

MAX_WEBTERM_ATTEMPTS: int = 10

_locks = {'webterm_thread': RLock()}

def get_webterm(state: WebState) -> Tuple[Any, Any]:
    """
    Start the webterm and return its iframe.
    """
    session_id = state['session-store.data'].get('session-id', None)
    username = active_sessions.get(session_id, {}).get('username', None)
    if not is_session_authenticated(session_id):
        msg = f"User '{username}' is not authorized to access the webterm."
        return (
            html.Div(
                html.Pre(msg, id='console-pre'),
                id = "console-div",
            ),
            [alert_from_success_tuple((
                False,
                "This Meerschaum instance only allows administrators to access the Webterm."
            ))]
        )

    for i in range(MAX_WEBTERM_ATTEMPTS):
        if is_webterm_running('localhost', 8765):
            return (
                html.Iframe(
                    src = f"/webterm?s={session_id}",
                    id = "webterm-iframe",
                ),
                []
            )
        time.sleep(1)
    return console_div, [alert_from_success_tuple((False, "Could not start the webterm server."))]



webterm_procs = {}
def start_webterm() -> None:
    """
    Start the webterm thread.
    """
    from meerschaum._internal.entry import entry
    from meerschaum.utils.packages import run_python_package

    def run():
        _ = run_python_package(
            'meerschaum',
            ['start', 'webterm'],
            capture_output = True,
            as_proc = True,
            store_proc_dict = webterm_procs,
            store_proc_key = 'process',
            venv = None,
        )

    with _locks['webterm_thread']:
        if webterm_procs.get('thread', None) is None:
            webterm_thread = Thread(target=run)
            webterm_procs['thread'] = webterm_thread
            webterm_thread.start()


def stop_webterm() -> None:
    """
    Stop the webterm thread.
    """
    webterm_thread = webterm_procs.get('thread', None)
    webterm_proc = webterm_procs.get('process', None)
    with _locks['webterm_thread']:
        if webterm_proc is not None:
            webterm_proc.terminate()
        if webterm_thread is not None:
            webterm_thread.join()



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Interact with Meerschaum APIs. May be chained together (see 'meerschaum:api_instance' in your config.yaml).
"""

from datetime import datetime, timedelta, timezone
from meerschaum.utils.typing import Optional
from meerschaum.connectors import Connector
from meerschaum.utils.warnings import warn, error
from meerschaum.utils.packages import attempt_import

required_attributes = {
    'host',
}

class APIConnector(Connector):
    """
    Connect to a Meerschaum API instance.
    """

    IS_INSTANCE: bool = True
    IS_THREAD_SAFE: bool = False

    from ._request import (
        make_request,
        get,
        post,
        put,
        patch,
        delete,
        wget,
    )
    from ._actions import get_actions, do_action
    from ._misc import get_mrsm_version, get_chaining_status
    from ._pipes import (
        register_pipe,
        fetch_pipes_keys,
        edit_pipe,
        sync_pipe,
        delete_pipe,
        get_pipe_data,
        get_pipe_id,
        get_pipe_attributes,
        get_sync_time,
        pipe_exists,
        create_metadata,
        get_pipe_rowcount,
        drop_pipe,
        clear_pipe,
        get_pipe_columns_types,
    )
    from ._fetch import fetch
    from ._plugins import (
        register_plugin,
        install_plugin,
        delete_plugin,
        get_plugins,
        get_plugin_attributes,
    )
    from ._login import login, test_connection
    from ._users import (
        register_user,
        get_user_id,
        get_users,
        edit_user,
        delete_user,
        get_user_password_hash,
        get_user_type,
        get_user_attributes,
    )
    from ._uri import from_uri

    def __init__(
        self,
        label: Optional[str] = None,
        wait: bool = False,
        debug: bool = False,
        **kw
    ):
        if 'uri' in kw:
            from_uri_params = self.from_uri(kw['uri'], as_dict=True)
            label = label or from_uri_params.get('label', None)
            from_uri_params.pop('label', None)
            kw.update(from_uri_params)

        super().__init__('api', label=label, **kw)
        if 'protocol' not in self.__dict__:
            self.protocol = 'http'
        if 'uri' not in self.__dict__:
            self.verify_attributes(required_attributes)
        else:
            from meerschaum.connectors.sql import SQLConnector
            conn_attrs = SQLConnector.parse_uri(self.__dict__['uri'])
            if 'host' not in conn_attrs:
                raise Exception(f"Invalid URI for '{self}'.")
            self.__dict__.update(conn_attrs)
        self.url = (
            self.protocol + '://' +
            self.host
            + (
                (':' + str(self.port))
                if self.__dict__.get('port', None)
                else ''
            )
        )
        self._token = None
        self._expires = None
        self._session = None


    @property
    def URI(self) -> str:
        """
        Return the fully qualified URI.
        """
        username = self.__dict__.get('username', None)
        password = self.__dict__.get('password', None)
        creds = (username + ':' + password + '@') if username and password else ''
        return (
            self.protocol
            + '://'
            + creds
            + self.host
            + (
                (':' + str(self.port))
                if self.__dict__.get('port', None)
                else ''
            )
        )


    @property
    def session(self):
        if self._session is None:
            requests = attempt_import('requests')
            if requests:
                self._session = requests.Session()
            if self._session is None:
                error(f"Failed to import requests. Is requests installed?")
        return self._session

    @property
    def token(self):
        expired = (
            True if self._expires is None else (
                (
                    self._expires
                    <
                    datetime.now(timezone.utc).replace(tzinfo=None) + timedelta(minutes=1)
                )
            )
        )

        if self._token is None or expired:
            success, msg = self.login()
            if not success:
                warn(msg, stack=False)
        return self._token



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Global pools that are joined on exit.
"""

from __future__ import annotations
from meerschaum.utils.typing import Optional, Callable, List, Any
from meerschaum.utils.threading import Lock, RLock
import signal

pools = {}
_locks = {
    'pools': Lock(),
}

def _initializer():
    """Ignore keyboard interrupt in workers."""
    signal.signal(signal.SIGINT, signal.SIG_IGN)

def get_pool(
        pool_class_name: str = 'ThreadPool',
        workers: Optional[int] = None,
        initializer: Optional[Callable[[None], None]] = None,
        initargs: Optional[List[Any]] = None,
    ):
    """If the requested pool does not exist, instantiate it here.
    Pools are joined and closed on exit."""
    from multiprocessing import cpu_count
    if workers is None:
        workers = cpu_count()
    pool_key = pool_class_name + f'-{workers}'

    def build_pool(workers):
        from meerschaum.utils.warnings import warn
        from meerschaum.utils.packages import attempt_import
        import importlib
        try:
            Pool = getattr(
                importlib.import_module('multiprocessing.pool'),
                pool_class_name
            )
        except Exception as e:
            warn(e, stacklevel=3)
            Pool = getattr(
                importlib.import_module('multiprocessing.pool'),
                'ThreadPool'
            )

        try:
            pool = Pool(workers, initializer=initializer, initargs=initargs)
        except Exception as e:
            print(e)
            pool = None

        with _locks['pools']:
            pools[pool_key] = pool

    if pools.get(pool_key, None) is None:
        build_pool(workers)

    if (
        pools[pool_key] is not None
        and pools[pool_key]._state not in ('RUN', 0)
    ):
        try:
            pools[pool_key].close()
            pools[pool_key].terminate()
        except Exception as e:
            pass
        del pools[pool_key]
        build_pool(workers)

    return pools[pool_key]


def get_pools():
    """Return the global pools dictionary."""
    global pools
    if pools is None:
        with _locks['pools']:
            pools = {}
    return pools


def get_pool_executor(workers: Optional[int] = None):
    """ Return a new `ThreadPoolExecutor`. """
    try:
        from multiprocessing import cpu_count
        from concurrent.futures import ThreadPoolExecutor
        workers = cpu_count() if workers is None else workers
    except Exception as e:
        return None

    return ThreadPoolExecutor(max_workers=workers) if ThreadPoolExecutor is not None else None



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Utility functions for working with data types.
"""

import traceback
from decimal import Decimal, Context, InvalidOperation
from meerschaum.utils.typing import Dict, Union, Any
from meerschaum.utils.warnings import warn

MRSM_PD_DTYPES: Dict[str, str] = {
    'json': 'object',
    'numeric': 'object',
    'datetime': 'datetime64[ns]',
    'bool': 'bool[pyarrow]',
    'int': 'Int64',
    'int8': 'Int8',
    'int16': 'Int16',
    'int32': 'Int32',
    'int64': 'Int64',
    'str': 'string[python]',
}


def to_pandas_dtype(dtype: str) -> str:
    """
    Cast a supported Meerschaum dtype to a Pandas dtype.
    """
    known_dtype = MRSM_PD_DTYPES.get(dtype, None)
    if known_dtype is not None:
        return known_dtype

    ### NOTE: Kind of a hack, but if the first word of the given dtype is in all caps,
    ### treat it as a SQL db type.
    if dtype.split(' ')[0].isupper():
        from meerschaum.utils.dtypes.sql import get_pd_type_from_db_type
        return get_pd_type_from_db_type(dtype)

    from meerschaum.utils.packages import attempt_import
    pandas = attempt_import('pandas', lazy=False)

    try:
        return str(pandas.api.types.pandas_dtype(dtype))
    except Exception as e:
        warn(
            f"Invalid dtype '{dtype}', will use 'object' instead:\n"
            + f"{traceback.format_exc()}",
            stack = False,
        )
    
    return 'object'


def are_dtypes_equal(
        ldtype: Union[str, Dict[str, str]],
        rdtype: Union[str, Dict[str, str]],
    ) -> bool:
    """
    Determine whether two dtype strings may be considered
    equivalent to avoid unnecessary conversions.

    Parameters
    ----------
    ldtype: Union[str, Dict[str, str]]
        The left dtype to compare.
        May also provide a dtypes dictionary.

    rdtype: Union[str, Dict[str, str]]
        The right dtype to compare.
        May also provide a dtypes dictionary.

    Returns
    -------
    A `bool` indicating whether the two dtypes are to be considered equivalent.
    """
    if isinstance(ldtype, dict) and isinstance(rdtype, dict):
        lkeys = sorted([str(k) for k in ldtype.keys()])
        rkeys = sorted([str(k) for k in rdtype.keys()])
        for lkey, rkey in zip(lkeys, rkeys):
            if lkey != rkey:
                return False
            ltype = ldtype[lkey]
            rtype = rdtype[rkey]
            if not are_dtypes_equal(ltype, rtype):
                return False
        return True

    try:
        if ldtype == rdtype:
            return True
    except Exception as e:
        warn(f"Exception when comparing dtypes, returning False:\n{traceback.format_exc()}")
        return False

    ### Sometimes pandas dtype objects are passed.
    ldtype = str(ldtype)
    rdtype = str(rdtype)

    json_dtypes = ('json', 'object')
    if ldtype in json_dtypes and rdtype in json_dtypes:
        return True

    numeric_dtypes = ('numeric', 'object')
    if ldtype in numeric_dtypes and rdtype in numeric_dtypes:
        return True

    ldtype_clean = ldtype.split('[')[0]
    rdtype_clean = rdtype.split('[')[0]

    if ldtype_clean.lower() == rdtype_clean.lower():
        return True

    datetime_dtypes = ('datetime', 'timestamp')
    ldtype_found_dt_prefix = False
    rdtype_found_dt_prefix = False
    for dt_prefix in datetime_dtypes:
        ldtype_found_dt_prefix = (dt_prefix in ldtype.lower()) or ldtype_found_dt_prefix
        rdtype_found_dt_prefix = (dt_prefix in rdtype.lower()) or rdtype_found_dt_prefix
    if ldtype_found_dt_prefix and rdtype_found_dt_prefix:
        return True

    string_dtypes = ('str', 'string', 'object')
    if ldtype_clean in string_dtypes and rdtype_clean in string_dtypes:
        return True

    int_dtypes = ('int', 'int64', 'int32', 'int16', 'int8')
    if ldtype_clean.lower() in int_dtypes and rdtype_clean.lower() in int_dtypes:
        return True

    float_dtypes = ('float', 'float64', 'float32', 'float16', 'float128', 'double')
    if ldtype_clean.lower() in float_dtypes and rdtype_clean.lower() in float_dtypes:
        return True

    bool_dtypes = ('bool', 'boolean')
    if ldtype_clean in bool_dtypes and rdtype_clean in bool_dtypes:
        return True

    return False


def is_dtype_numeric(dtype: str) -> bool:
    """
    Determine whether a given `dtype` string
    should be considered compatible with the Meerschaum dtype `numeric`.

    Parameters
    ----------
    dtype: str
        The pandas-like dtype string.

    Returns
    -------
    A bool indicating the dtype is compatible with `numeric`.
    """
    dtype_lower = dtype.lower()

    acceptable_substrings = ('numeric', 'float', 'double', 'int')
    for substring in acceptable_substrings:
        if substring in dtype_lower:
            return True

    return False


def attempt_cast_to_numeric(value: Any) -> Any:
    """
    Given a value, attempt to coerce it into a numeric (Decimal).
    """
    if isinstance(value, Decimal):
        return value
    try:
        return (
            Decimal(str(value))
            if not value_is_null(value)
            else Decimal('NaN')
        )
    except Exception as e:
        return value


def value_is_null(value: Any) -> bool:
    """
    Determine if a value is a null-like string.
    """
    return str(value).lower() in ('none', 'nan', 'na', 'nat', '', '<na>')


def none_if_null(value: Any) -> Any:
    """
    Return `None` if a value is a null-like string.
    """
    return (None if value_is_null(value) else value)


def quantize_decimal(x: Decimal, scale: int, precision: int) -> Decimal:
    """
    Quantize a given `Decimal` to a known scale and precision.

    Parameters
    ----------
    x: Decimal
        The `Decimal` to be quantized.

    scale: int
        The total number of significant digits.

    precision: int
        The number of significant digits after the decimal point.

    Returns
    -------
    A `Decimal` quantized to the specified scale and precision.
    """
    precision_decimal = Decimal((('1' * scale) + '.' + ('1' * precision)))
    try:
        return x.quantize(precision_decimal, context=Context(prec=scale))
    except InvalidOperation:
        return x



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Utility functions for parsing connector keys.
"""

from __future__ import annotations
from meerschaum.utils.typing import Mapping, Any, SuccessTuple, Union, Optional, Dict, Tuple

def parse_connector_keys(
        keys: str,
        construct: bool = True,
        as_tuple: bool = False,
        **kw:  Any
    ) -> (
        Union[
            meerschaum.connectors.Connector,
            Dict[str, Any],
            Tuple[
                Union[
                    meerschaum.connectors.Connector,
                    Dict[str, Any],
                    None
                ],
                str
            ],
            None
        ]
    ):
    """
    Convenience function for parsing connector keys and returning Connector objects.

    Parameters
    ----------
    keys: str
        Keys are split by a colon (`':'`) into type and label. If the label is omitted,
        (e.g. 'sql'), pass it along to `meerschaum.connectors.get_connector` to parse.

    construct: bool, default True
        If True, return a Connector. Otherwise return the configuration dictionary for a Connector.
        **NOTE:** This may include passwords, so be careful.

    as_tuple: bool, default False
        If True, return a tuple of (conn, keys). `conn` may be a dict or connector.

    Returns
    -------
    A connector or dictionary of attributes. If `as_tuple`, also return the connector's keys.

    """
    import copy
    from meerschaum.connectors import get_connector
    from meerschaum.config import get_config
    from meerschaum.config.static import STATIC_CONFIG
    from meerschaum.utils.warnings import error

    ### `get_connector()` handles the logic for falling back to 'main',
    ### so don't make any decisions here.
    vals = str(keys).split(':')
    _type = vals[0]
    _label = vals[1] if len(vals) > 1 else STATIC_CONFIG['connectors']['default_label']
    _get_connector_kw = {'type': _type, 'label': _label}
    _get_connector_kw.update(kw)

    if construct:
        conn = get_connector(**_get_connector_kw)
        if conn is None:
            error(f"Unable to parse connector keys '{keys}'", stack=False)
    else:
        connectors_config = get_config('meerschaum', 'connectors', patch=True)
        ### invalid type
        if vals[0] not in connectors_config:
            return None
        type_config = get_config('meerschaum', 'connectors', _type)

        default_config = copy.deepcopy(type_config.get('default', {}))
        conn = type_config.get(_label, None)
        if default_config is not None and conn is not None:
            default_config.update(conn)
            conn = default_config

    if as_tuple:
        return conn, _type + ':' + _label
    return conn


def parse_instance_keys(
        keys: Optional[str],
        construct: bool = True,
        as_tuple: bool = False,
        **kw
    ):
    """
    Parse the Meerschaum instance value into a Connector object.
    """
    from meerschaum.utils.warnings import warn
    from meerschaum.config import get_config

    ### TODO Check for valid types? Not sure how to do that if construct = False.

    if keys is None:
        keys = get_config('meerschaum', 'instance')
    keys = str(keys)
    
    return parse_connector_keys(keys, construct=construct, as_tuple=as_tuple, **kw)


def parse_repo_keys(keys: Optional[str] = None, **kw):
    """Parse the Meerschaum repository value into an APIConnector."""
    from meerschaum.config import get_config
    if keys is None:
        keys = get_config('meerschaum', 'default_repository', patch=True)
    keys = str(keys)
    if ':' not in keys:
        keys = 'api:' + keys

    return parse_connector_keys(keys, **kw)


def is_valid_connector_keys(
        keys: str
    ) -> bool:
    """Verify a connector_keys string references a valid connector.
    """
    try:
        success = parse_connector_keys(keys, construct=False) is not None
    except Exception as e:
        success = False
    return success




#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Functions for interacting with jobs via the web interface.
"""

from __future__ import annotations
from dash.dependencies import Input, Output, State
from meerschaum.utils.typing import List, Optional, Dict, Any, Tuple, Union, WebState
from meerschaum.utils.packages import attempt_import, import_html, import_dcc
from meerschaum.api.dash.components import alert_from_success_tuple, build_cards_grid
from meerschaum.api.dash.users import is_session_authenticated
from meerschaum.api import CHECK_UPDATE
dbc = attempt_import('dash_bootstrap_components', lazy=False, check_update=CHECK_UPDATE)
html, dcc = import_html(), import_dcc()
dateutil_parser = attempt_import('dateutil.parser', check_update=CHECK_UPDATE)
from meerschaum.utils.daemon import (
    get_daemons,
    get_running_daemons,
    get_paused_daemons,
    get_stopped_daemons,
    Daemon,
)
from meerschaum.config import get_config
from meerschaum.utils.misc import sorted_dict

STATUS_EMOJI: Dict[str, str] = {
    'running': get_config('formatting', 'emoji', 'running'),
    'paused': get_config('formatting', 'emoji', 'paused'),
    'stopped': get_config('formatting', 'emoji', 'stopped'),
    'dne': get_config('formatting', 'emoji', 'failure')
}

def get_jobs_cards(state: WebState):
    """
    Build cards and alerts lists for jobs.
    """
    daemons = get_daemons()
    session_id = state['session-store.data'].get('session-id', None)
    is_authenticated = is_session_authenticated(session_id)

    alert = alert_from_success_tuple(daemons)
    cards = []

    for d in daemons:
        footer_children = html.Div(
            build_process_timestamps_children(d),
            id = {'type': 'process-timestamps-div', 'index': d.daemon_id},
        )
        follow_logs_button = dbc.DropdownMenuItem(
            "Follow logs",
            id = {'type': 'follow-logs-button', 'index': d.daemon_id},
        )
        download_logs_button = dbc.DropdownMenuItem(
            "Download logs",
            id = {'type': 'job-download-logs-button', 'index': d.daemon_id},
        )
        logs_menu_children = (
            ([follow_logs_button] if is_authenticated else []) + [download_logs_button]
        )
        header_children = [
            html.Div(
                build_status_children(d),
                id = {'type': 'manage-job-status-div', 'index': d.daemon_id},
                style = {'float': 'left'},
            ),
            html.Div(
                dbc.DropdownMenu(
                    logs_menu_children,
                    label = "Logs",
                    size = "sm",
                    align_end = True,
                    color = "secondary",
                    menu_variant = 'dark',
                ),
                style = {'float': 'right'},
            ),
        ]

        body_children = [
            html.H4(html.B(d.daemon_id), className="card-title"),
            html.Div(
                html.P(
                    d.label,
                    className = "card-text job-card-text", 
                    style = {"word-wrap": "break-word"}
                ),
                style = {"white-space": "pre-wrap"},
            ),
            html.Div(
                (
                    build_manage_job_buttons_div_children(d)
                    if is_authenticated
                    else []
                ),
                id = {'type': 'manage-job-buttons-div', 'index': d.daemon_id},
            ),
            html.Div(id={'type': 'manage-job-alert-div', 'index': d.daemon_id}),
        ]

        cards.append(
            dbc.Card([
                dbc.CardHeader(header_children),
                dbc.CardBody(body_children),
                dbc.CardFooter(footer_children),
            ])
        )

    return cards, []


def build_manage_job_buttons_div_children(daemon: Daemon):
    """
    Return the children for the manage job buttons div.
    """
    buttons = build_manage_job_buttons(daemon)
    if not buttons:
        return []
    return [
        html.Br(),
        dbc.Row([
            dbc.Col(button, width=6)
            for button in buttons
        ])
    ]


def build_manage_job_buttons(daemon: Daemon):
    """
    Return the currently available job management buttons for a given Daemon.
    """
    if daemon is None:
        return []
    start_button = dbc.Button(
        'Start',
        size = 'sm',
        color = 'success',
        style = {'width': '100%'},
        id = {
            'type': 'manage-job-button',
            'action': 'start',
            'index': daemon.daemon_id,
        },
    )
    pause_button = dbc.Button(
        'Pause',
        size = 'sm',
        color = 'warning',
        style = {'width': '100%'},
        id = {
            'type': 'manage-job-button',
            'action': 'pause',
            'index': daemon.daemon_id,
        },
    )
    stop_button = dbc.Button(
        'Stop',
        size = 'sm',
        color = 'danger',
        style = {'width': '100%'},
        id = {
            'type': 'manage-job-button',
            'action': 'stop',
            'index': daemon.daemon_id,
        },
    )
    delete_button = dbc.Button(
        'Delete',
        size = 'sm',
        color = 'danger',
        style = {'width': '100%'},
        id = {
            'type': 'manage-job-button',
            'action': 'delete',
            'index': daemon.daemon_id,
        },
    )
    buttons = []
    if daemon.status in ('stopped', 'paused'):
        buttons.append(start_button)
    if daemon.status == 'stopped':
        buttons.append(delete_button)
    if daemon.status in ('running',):
        buttons.append(pause_button)
    if daemon.status in ('running', 'paused'):
        buttons.append(stop_button)

    return buttons


def build_status_children(daemon: Daemon) -> List[html.P]:
    """
    Return the status HTML component for this daemon.
    """
    if daemon is None:
        return STATUS_EMOJI['dne']

    status_str = (
        STATUS_EMOJI.get(daemon.status, STATUS_EMOJI['stopped'])
        + ' '
        + daemon.status.capitalize()
    )
    return html.P(
        html.B(status_str),
        className = f"{daemon.status}-job",
    )


def build_process_timestamps_children(daemon: Daemon) -> List[dbc.Row]:
    """
    Return the children to the process timestamps in the footer of the job card.
    """
    if daemon is None:
        return []
    children = []
    for timestamp_key, timestamp_val in sorted_dict(
        daemon.properties.get('process', {})
    ).items():
        timestamp = dateutil_parser.parse(timestamp_val)
        timestamp_str = timestamp.strftime('%Y-%m-%d %H:%M UTC')
        children.append(
            dbc.Row(
                [
                    dbc.Col(
                        html.P(
                            timestamp_key.capitalize(),
                            style = {'font-size': 'small'},
                            className = 'text-muted mb-0',
                        ),
                        width = 4,
                    ),
                    dbc.Col(
                        html.P(
                            timestamp_str,
                            style = {'font-size': 'small', 'text-align': 'right'},
                            className = 'text-muted mb-0',
                        ),
                        width = 8,
                    ),
                ],
                justify = 'between', 
            )
        )
    return children



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Define utilities for instance connectors.
"""

import time
from datetime import datetime, timezone, timedelta
import meerschaum as mrsm
from meerschaum.utils.typing import Dict, SuccessTuple, Optional, Union, List
from meerschaum.utils.warnings import warn


_in_memory_temp_tables: Dict[str, bool] = {}
def _log_temporary_tables_creation(
        self,
        tables: Union[str, List[str]],
        ready_to_drop: bool = False,
        create: bool = True,
        debug: bool = False,
    ) -> SuccessTuple:
    """
    Log a temporary table's creation for later deletion.
    """
    from meerschaum.utils.misc import items_str
    from meerschaum.connectors.sql.tables import get_tables
    sqlalchemy = mrsm.attempt_import('sqlalchemy')
    temp_tables_table = get_tables(
        mrsm_instance = self,
        create = create,
        debug = debug,
    )['temp_tables']
    if isinstance(tables, str):
        tables = [tables]
    now = datetime.now(timezone.utc).replace(tzinfo=None)
    docs = [
        {
            'date_created': now,
            'table': table,
            'ready_to_drop': (now if ready_to_drop else None),
        }
        for table in tables
    ]
    ### NOTE: We may be running in a temporary context, in which we don't create instance tables.
    queries = [sqlalchemy.insert(temp_tables_table).values(**doc) for doc in docs]
    results = [self.exec(query, silent=True, debug=debug) for query in queries]
    success = all(results)
    _in_memory_temp_tables.update(
        {
            table: ready_to_drop
            for table in tables
        }
    )
    msg = "Success" if success else f"Failed to log temp tables {items_str(tables)}."
    return success, msg


def _drop_temporary_table(
        self,
        table: str,
        debug: bool = False,
    ) -> SuccessTuple:
    """
    Drop a temporary table and clear it from the internal table.
    """
    from meerschaum.utils.sql import sql_item_name, table_exists, SKIP_IF_EXISTS_FLAVORS
    if_exists = "IF EXISTS" if self.flavor not in SKIP_IF_EXISTS_FLAVORS else ""
    if not if_exists:
        if not table_exists(table, self, self.internal_schema, debug=debug):
            return True, "Success"

    drop_query = f"DROP TABLE {if_exists} " + sql_item_name(
        table, self.flavor, schema=self.internal_schema
    )
    drop_success = self.exec(drop_query, silent=True, debug=debug) is not None
    drop_msg = "Success" if drop_success else f"Failed to drop temporary table '{table}'."
    return drop_success, drop_msg


def _drop_temporary_tables(self, debug: bool = False) -> SuccessTuple:
    """
    Drop all tables in the internal schema that are marked as ready to be dropped.
    """
    from meerschaum.utils.sql import sql_item_name, table_exists
    from meerschaum.utils.misc import items_str
    from meerschaum.connectors.sql.tables import get_tables
    sqlalchemy = mrsm.attempt_import('sqlalchemy')
    temp_tables_table = get_tables(
        mrsm_instance = self,
        create = False,
        debug = debug,
    )['temp_tables']
    query = (
        sqlalchemy.select(temp_tables_table.c.table)
        .where(temp_tables_table.c.ready_to_drop.is_not(None))
    )
    tables_to_drop = [
        table
        for table, ready_to_drop in _in_memory_temp_tables.items()
        if ready_to_drop
    ]
    if not tables_to_drop:
        df = self.read(query, silent=True, debug=debug)
        tables_to_drop = (
            list(df['table'])
            if df is not None
            else []
        )
    if not tables_to_drop:
        return True, "Success"

    dropped_tables = []
    failed_tables = []
    for table in tables_to_drop:
        drop_success, drop_msg = self._drop_temporary_table(table, debug=debug)
        if not drop_success:
            failed_tables.append(table)
            continue
        dropped_tables.append(table)
        _ = _in_memory_temp_tables.pop(table, None)

    if dropped_tables:
        delete_query = (
            sqlalchemy.delete(temp_tables_table)
            .where(temp_tables_table.c.table.in_(dropped_tables))
        )
        delete_result = self.exec(delete_query, silent=True, debug=debug)

    success = len(failed_tables) == 0
    msg = (
        "Success"
        if success
        else (
            "Failed to drop temporary tables "
            + f"{items_str(failed_tables)}."
        )
    )
    return success, msg


def _drop_old_temporary_tables(
        self,
        refresh: bool = True,
        debug: bool = False,
    ) -> SuccessTuple:
    """
    Drop temporary tables older than the configured interval (24 hours by default).
    """
    from meerschaum.config import get_config
    from meerschaum.connectors.sql.tables import get_tables
    from meerschaum.utils.misc import items_str
    sqlalchemy = mrsm.attempt_import('sqlalchemy')
    temp_tables_table = get_tables(mrsm_instance=self, create=False, debug=debug)['temp_tables']
    last_check = getattr(self, '_stale_temporary_tables_check_timestamp', 0)
    now_ts = time.perf_counter()
    if refresh or not last_check or (now_ts - last_check) > 60:
        self._stale_temporary_tables_check_timestamp = now_ts
        return self._drop_temporary_tables(debug=debug)

    stale_temporary_tables_minutes = get_config(
        'system', 'connectors', 'sql', 'instance', 'stale_temporary_tables_minutes'
    )
    now = datetime.now(timezone.utc).replace(tzinfo=None)
    end = now - timedelta(minutes=stale_temporary_tables_minutes)

    query = (
        sqlalchemy.select(temp_tables_table.c.table)
        .where(temp_tables_table.c.date_created < end)
    )

    df = self.read(query, silent=True, debug=debug)
    if df is None:
        return True, "Success"

    ### Insert new records with the current time (skipping updates to avoid recursion).
    docs = [
        {
            'date_created': now,
            'table': table,
            'ready_to_drop': now,
        }
        for table in df['table']
    ]
    if docs:
        queries = [sqlalchemy.insert(temp_tables_table).values(**doc) for doc in docs]
        results = [self.exec(query, silent=True, debug=debug) for query in queries]
        _in_memory_temp_tables.update(
            {
                table: True
                for table in df['table']
            }
        )

    return self._drop_temporary_tables(debug=debug)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
"""
This module contains SQLConnector functions for executing SQL queries.
"""

from __future__ import annotations
from meerschaum.utils.typing import (
    Union, Mapping, List, Dict, SuccessTuple, Optional, Any, Iterable, Callable,
    Tuple, Hashable,
)

from meerschaum.utils.debug import dprint
from meerschaum.utils.warnings import warn

### database flavors that can use bulk insert
_bulk_flavors = {'postgresql', 'timescaledb', 'citus'}
### flavors that do not support chunks
_disallow_chunks_flavors = {'duckdb', 'mssql'}
_max_chunks_flavors = {'sqlite': 1000,}

def read(
        self,
        query_or_table: Union[str, sqlalchemy.Query],
        params: Union[Dict[str, Any], List[str], None] = None,
        dtype: Optional[Dict[str, Any]] = None,
        coerce_float: bool = True,
        chunksize: Optional[int] = -1,
        workers: Optional[int] = None,
        chunk_hook: Optional[Callable[[pandas.DataFrame], Any]] = None,
        as_hook_results: bool = False,
        chunks: Optional[int] = None,
        schema: Optional[str] = None,
        as_chunks: bool = False,
        as_iterator: bool = False,
        as_dask: bool = False,
        index_col: Optional[str] = None,
        silent: bool = False,
        debug: bool = False,
        **kw: Any
    ) -> Union[
        pandas.DataFrame,
        dask.DataFrame,
        List[pandas.DataFrame],
        List[Any],
        None,
    ]:
    """
    Read a SQL query or table into a pandas dataframe.

    Parameters
    ----------
    query_or_table: Union[str, sqlalchemy.Query]
        The SQL query (sqlalchemy Query or string) or name of the table from which to select.

    params: Optional[Dict[str, Any]], default None
        `List` or `Dict` of parameters to pass to `pandas.read_sql()`.
        See the pandas documentation for more information:
        https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.read_sql.html

    dtype: Optional[Dict[str, Any]], default None
        A dictionary of data types to pass to `pandas.read_sql()`.
        See the pandas documentation for more information:
        https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.read_sql_query.html

    chunksize: Optional[int], default -1
        How many chunks to read at a time. `None` will read everything in one large chunk.
        Defaults to system configuration.

        **NOTE:** DuckDB does not allow for chunking.

    workers: Optional[int], default None
        How many threads to use when consuming the generator.
        Only applies if `chunk_hook` is provided.

    chunk_hook: Optional[Callable[[pandas.DataFrame], Any]], default None
        Hook function to execute once per chunk, e.g. writing and reading chunks intermittently.
        See `--sync-chunks` for an example.
        **NOTE:** `as_iterator` MUST be False (default).

    as_hook_results: bool, default False
        If `True`, return a `List` of the outputs of the hook function.
        Only applicable if `chunk_hook` is not None.

        **NOTE:** `as_iterator` MUST be `False` (default).

    chunks: Optional[int], default None
        Limit the number of chunks to read into memory, i.e. how many chunks to retrieve and
        return into a single dataframe.
        For example, to limit the returned dataframe to 100,000 rows,
        you could specify a `chunksize` of `1000` and `chunks` of `100`.

    schema: Optional[str], default None
        If just a table name is provided, optionally specify the table schema.
        Defaults to `SQLConnector.schema`.

    as_chunks: bool, default False
        If `True`, return a list of DataFrames. 
        Otherwise return a single DataFrame.

    as_iterator: bool, default False
        If `True`, return the pandas DataFrame iterator.
        `chunksize` must not be `None` (falls back to 1000 if so),
        and hooks are not called in this case.

    index_col: Optional[str], default None
        If using Dask, use this column as the index column.
        If omitted, a Pandas DataFrame will be fetched and converted to a Dask DataFrame.

    silent: bool, default False
        If `True`, don't raise warnings in case of errors.
        Defaults to `False`.

    Returns
    -------
    A `pd.DataFrame` (default case), or an iterator, or a list of dataframes / iterators,
    or `None` if something breaks.

    """
    if chunks is not None and chunks <= 0:
        return []
    from meerschaum.utils.sql import sql_item_name, truncate_item_name
    from meerschaum.utils.dtypes.sql import NUMERIC_PRECISION_FLAVORS
    from meerschaum.utils.packages import attempt_import, import_pandas
    from meerschaum.utils.pool import get_pool
    from meerschaum.utils.dataframe import chunksize_to_npartitions, get_numeric_cols
    import warnings
    import inspect
    import traceback
    from decimal import Decimal
    pd = import_pandas()
    dd = None
    is_dask = 'dask' in pd.__name__
    pd = attempt_import('pandas')
    is_dask = dd is not None
    npartitions = chunksize_to_npartitions(chunksize)
    if is_dask:
        chunksize = None
    schema = schema or self.schema

    sqlalchemy = attempt_import("sqlalchemy")
    default_chunksize = self._sys_config.get('chunksize', None)
    chunksize = chunksize if chunksize != -1 else default_chunksize
    if chunksize is None and as_iterator:
        if not silent and self.flavor not in _disallow_chunks_flavors:
            warn(
                f"An iterator may only be generated if chunksize is not None.\n"
                + "Falling back to a chunksize of 1000.", stacklevel=3,
            )
        chunksize = 1000
    if chunksize is not None and self.flavor in _max_chunks_flavors:
        if chunksize > _max_chunks_flavors[self.flavor]:
            if chunksize != default_chunksize:
                warn(
                    f"The specified chunksize of {chunksize} exceeds the maximum of "
                    + f"{_max_chunks_flavors[self.flavor]} for flavor '{self.flavor}'.\n"
                    + f"    Falling back to a chunksize of {_max_chunks_flavors[self.flavor]}.",
                    stacklevel = 3,
                )
            chunksize = _max_chunks_flavors[self.flavor]

    ### NOTE: A bug in duckdb_engine does not allow for chunks.
    if chunksize is not None and self.flavor in _disallow_chunks_flavors:
        chunksize = None

    if debug:
        import time
        start = time.perf_counter()
        dprint(f"[{self}]\n{query_or_table}")
        dprint(f"[{self}] Fetching with chunksize: {chunksize}")

    ### This might be sqlalchemy object or the string of a table name.
    ### We check for spaces and quotes to see if it might be a weird table.
    if (
        ' ' not in str(query_or_table)
        or (
            ' ' in str(query_or_table)
            and str(query_or_table).startswith('"')
            and str(query_or_table).endswith('"')
        )
    ):
        truncated_table_name = truncate_item_name(str(query_or_table), self.flavor)
        if truncated_table_name != str(query_or_table) and not silent:
            warn(
                f"Table '{name}' is too long for '{self.flavor}',"
                + f" will instead create the table '{truncated_name}'."
            )

        query_or_table = sql_item_name(str(query_or_table), self.flavor, schema)
        if debug:
            dprint(f"[{self}] Reading from table {query_or_table}")
        formatted_query = sqlalchemy.text("SELECT * FROM " + str(query_or_table))
        str_query = f"SELECT * FROM {query_or_table}"
    else:
        str_query = query_or_table

    formatted_query = (
        sqlalchemy.text(str_query)
        if not is_dask and isinstance(str_query, str)
        else format_sql_query_for_dask(str_query)
    )

    chunk_list = []
    chunk_hook_results = []
    try:
        stream_results = not as_iterator and chunk_hook is not None and chunksize is not None
        with warnings.catch_warnings():
            warnings.filterwarnings('ignore', 'case sensitivity issues')

            read_sql_query_kwargs = {
                'params': params,
                'dtype': dtype,
                'coerce_float': coerce_float,
                'index_col': index_col,
            }
            if is_dask:
                if index_col is None:
                    dd = None
                    pd = attempt_import('pandas')
                    read_sql_query_kwargs.update({
                        'chunksize': chunksize,
                    })
            else:
                read_sql_query_kwargs.update({
                    'chunksize': chunksize,
                })

            if is_dask and dd is not None:
                ddf = dd.read_sql_query(
                    formatted_query,
                    self.URI,
                    **read_sql_query_kwargs
                )
            else:

                with self.engine.begin() as transaction:
                    with transaction.execution_options(stream_results=stream_results) as connection:
                        chunk_generator = pd.read_sql_query(
                            formatted_query,
                            connection,
                            **read_sql_query_kwargs
                        )

                        ### `stream_results` must be False (will load everything into memory).
                        if as_iterator or chunksize is None:
                            return chunk_generator

                        ### We must consume the generator in this context if using server-side cursors.
                        if stream_results:

                            pool = get_pool(workers=workers)

                            def _process_chunk(_chunk, _retry_on_failure: bool = True):
                                if not as_hook_results:
                                    chunk_list.append(_chunk)
                                result = None
                                if chunk_hook is not None:
                                    try:
                                        result = chunk_hook(
                                            _chunk,
                                            workers = workers,
                                            chunksize = chunksize,
                                            debug = debug,
                                            **kw
                                        )
                                    except Exception as e:
                                        result = False, traceback.format_exc()
                                        from meerschaum.utils.formatting import get_console
                                        if not silent:
                                            get_console().print_exception()

                                    ### If the chunk fails to process, try it again one more time.
                                    if isinstance(result, tuple) and result[0] is False:
                                        if _retry_on_failure:
                                            return _process_chunk(_chunk, _retry_on_failure=False)

                                return result

                            chunk_hook_results = list(pool.imap(_process_chunk, chunk_generator))
                            if as_hook_results:
                                return chunk_hook_results

    except Exception as e:
        if debug:
            dprint(f"[{self}] Failed to execute query:\n\n{query_or_table}\n\n")
        if not silent:
            warn(str(e), stacklevel=3)
        from meerschaum.utils.formatting import get_console
        if not silent:
            get_console().print_exception()

        return None

    if is_dask and dd is not None:
        ddf = ddf.reset_index()
        return ddf

    chunk_list = []
    read_chunks = 0
    chunk_hook_results = []
    if chunksize is None:
        chunk_list.append(chunk_generator)
    elif as_iterator:
        return chunk_generator
    else:
        try:
            for chunk in chunk_generator:
                if chunk_hook is not None:
                    chunk_hook_results.append(
                        chunk_hook(chunk, chunksize=chunksize, debug=debug, **kw)
                    )
                chunk_list.append(chunk)
                read_chunks += 1
                if chunks is not None and read_chunks >= chunks:
                    break
        except Exception as e:
            warn(f"[{self}] Failed to retrieve query results:\n" + str(e), stacklevel=3)
            from meerschaum.utils.formatting import get_console
            if not silent:
                get_console().print_exception()

    read_chunks = 0
    try:
        for chunk in chunk_generator:
            if chunk_hook is not None:
                chunk_hook_results.append(
                    chunk_hook(chunk, chunksize=chunksize, debug=debug, **kw)
                )
            chunk_list.append(chunk)
            read_chunks += 1
            if chunks is not None and read_chunks >= chunks:
                break
    except Exception as e:
        warn(f"[{self}] Failed to retrieve query results:\n" + str(e), stacklevel=3)
        from meerschaum.utils.formatting import get_console
        if not silent:
            get_console().print_exception()

        return None

    ### If no chunks returned, read without chunks
    ### to get columns
    if len(chunk_list) == 0:
        with warnings.catch_warnings():
            warnings.filterwarnings('ignore', 'case sensitivity issues')
            _ = read_sql_query_kwargs.pop('chunksize', None)
            with self.engine.begin() as connection:
                chunk_list.append(
                    pd.read_sql_query(
                        formatted_query,
                        connection,
                        **read_sql_query_kwargs
                    )
                )

    ### call the hook on any missed chunks.
    if chunk_hook is not None and len(chunk_list) > len(chunk_hook_results):
        for c in chunk_list[len(chunk_hook_results):]:
            chunk_hook_results.append(
                chunk_hook(c, chunksize=chunksize, debug=debug, **kw)
            )

    ### chunksize is not None so must iterate
    if debug:
        end = time.perf_counter()
        dprint(f"Fetched {len(chunk_list)} chunks in {round(end - start, 2)} seconds.")

    if as_hook_results:
        return chunk_hook_results
    
    ### Skip `pd.concat()` if `as_chunks` is specified.
    if as_chunks:
        for c in chunk_list:
            c.reset_index(drop=True, inplace=True)
            for col in get_numeric_cols(c):
                c[col] = c[col].apply(lambda x: x.canonical() if isinstance(x, Decimal) else x)
        return chunk_list

    df = pd.concat(chunk_list).reset_index(drop=True)
    ### NOTE: The calls to `canonical()` are to drop leading and trailing zeroes.
    for col in get_numeric_cols(df):
        df[col] = df[col].apply(lambda x: x.canonical() if isinstance(x, Decimal) else x)

    return df


def value(
        self,
        query: str,
        *args: Any,
        use_pandas: bool = False,
        **kw: Any
    ) -> Any:
    """
    Execute the provided query and return the first value.

    Parameters
    ----------
    query: str
        The SQL query to execute.
        
    *args: Any
        The arguments passed to `meerschaum.connectors.sql.SQLConnector.exec`
        if `use_pandas` is `False` (default) or to `meerschaum.connectors.sql.SQLConnector.read`.
        
    use_pandas: bool, default False
        If `True`, use `meerschaum.connectors.SQLConnector.read`, otherwise use
        `meerschaum.connectors.sql.SQLConnector.exec` (default).
        **NOTE:** This is always `True` for DuckDB.

    **kw: Any
        See `args`.

    Returns
    -------
    Any value returned from the query.

    """
    from meerschaum.utils.packages import attempt_import
    sqlalchemy = attempt_import('sqlalchemy')
    if self.flavor == 'duckdb':
        use_pandas = True
    if use_pandas:
        try:
            return self.read(query, *args, **kw).iloc[0, 0]
        except Exception as e:
            return None

    _close = kw.get('close', True)
    _commit = kw.get('commit', (self.flavor != 'mssql'))
    try:
        result, connection = self.exec(
            query,
            *args,
            with_connection = True,
            close = False,
            commit = _commit,
            **kw
        )
        first = result.first() if result is not None else None
        _val = first[0] if first is not None else None
    except Exception as e:
        warn(e, stacklevel=3)
        return None
    if _close:
        try:
            connection.close()
        except Exception as e:
            warn("Failed to close connection with exception:\n" + str(e))
    return _val


def execute(
        self,
        *args : Any,
        **kw : Any
    ) -> Optional[sqlalchemy.engine.result.resultProxy]:
    """
    An alias for `meerschaum.connectors.sql.SQLConnector.exec`.
    """
    return self.exec(*args, **kw)


def exec(
        self,
        query: str,
        *args: Any,
        silent: bool = False,
        debug: bool = False,
        commit: Optional[bool] = None,
        close: Optional[bool] = None,
        with_connection: bool = False,
        **kw: Any
    ) -> Union[
            sqlalchemy.engine.result.resultProxy,
            sqlalchemy.engine.cursor.LegacyCursorResult,
            Tuple[sqlalchemy.engine.result.resultProxy, sqlalchemy.engine.base.Connection],
            Tuple[sqlalchemy.engine.cursor.LegacyCursorResult, sqlalchemy.engine.base.Connection],
            None
    ]:
    """
    Execute SQL code and return the `sqlalchemy` result, e.g. when calling stored procedures.
    
    If inserting data, please use bind variables to avoid SQL injection!

    Parameters
    ----------
    query: Union[str, List[str], Tuple[str]]
        The query to execute.
        If `query` is a list or tuple, call `self.exec_queries()` instead.

    args: Any
        Arguments passed to `sqlalchemy.engine.execute`.
        
    silent: bool, default False
        If `True`, suppress warnings.

    commit: Optional[bool], default None
        If `True`, commit the changes after execution.
        Causes issues with flavors like `'mssql'`.
        This does not apply if `query` is a list of strings.

    close: Optional[bool], default None
        If `True`, close the connection after execution.
        Causes issues with flavors like `'mssql'`.
        This does not apply if `query` is a list of strings.

    with_connection: bool, default False
        If `True`, return a tuple including the connection object.
        This does not apply if `query` is a list of strings.
    
    Returns
    -------
    The `sqlalchemy` result object, or a tuple with the connection if `with_connection` is provided.

    """
    if isinstance(query, (list, tuple)):
        return self.exec_queries(
            list(query),
            *args,
            silent = silent,
            debug = debug,
            **kw
        )

    from meerschaum.utils.packages import attempt_import
    sqlalchemy = attempt_import("sqlalchemy")
    if debug:
        dprint(f"[{self}] Executing query:\n{query}")

    _close = close if close is not None else (self.flavor != 'mssql')
    _commit = commit if commit is not None else (
        (self.flavor != 'mssql' or 'select' not in str(query).lower())
    )

    ### Select and Insert objects need to be compiled (SQLAlchemy 2.0.0+).
    if not hasattr(query, 'compile'):
        query = sqlalchemy.text(query)

    connection = self.engine.connect()
    transaction = connection.begin() if _commit else None
    try:
        result = connection.execute(query, *args, **kw)
        if _commit:
            transaction.commit()
    except Exception as e:
        if debug:
            dprint(f"[{self}] Failed to execute query:\n\n{query}\n\n{e}")
        if not silent:
            warn(str(e), stacklevel=3)
        result = None
        if _commit:
            transaction.rollback()
    finally:
        if _close:
            connection.close()

        if with_connection:
            return result, connection

    return result


def exec_queries(
        self,
        queries: List[
            Union[
                str,
                Tuple[str, Callable[['sqlalchemy.orm.session.Session'], List[str]]]
            ]
        ],
        break_on_error: bool = False,
        rollback: bool = True,
        silent: bool = False,
        debug: bool = False,
    ) -> List[sqlalchemy.engine.cursor.LegacyCursorResult]:
    """
    Execute a list of queries in a single transaction.

    Parameters
    ----------
    queries: List[
        Union[
            str,
            Tuple[str, Callable[[], List[str]]]
        ]
    ]
        The queries in the transaction to be executed.
        If a query is a tuple, the second item of the tuple
        will be considered a callable hook that returns a list of queries to be executed
        before the next item in the list.

    break_on_error: bool, default True
        If `True`, stop executing when a query fails.

    rollback: bool, default True
        If `break_on_error` is `True`, rollback the transaction if a query fails.

    silent: bool, default False
        If `True`, suppress warnings.

    Returns
    -------
    A list of SQLAlchemy results.
    """
    from meerschaum.utils.warnings import warn
    from meerschaum.utils.debug import dprint
    from meerschaum.utils.packages import attempt_import
    sqlalchemy, sqlalchemy_orm = attempt_import('sqlalchemy', 'sqlalchemy.orm')
    session = sqlalchemy_orm.Session(self.engine)

    result = None
    results = []
    with session.begin():
        for query in queries:
            hook = None
            result = None

            if isinstance(query, tuple):
                query, hook = query
            if isinstance(query, str):
                query = sqlalchemy.text(query)

            if debug:
                dprint(f"[{self}]\n" + str(query))
 
            try:
                result = session.execute(query)
                session.flush()
            except Exception as e:
                msg = (f"Encountered error while executing:\n{e}")
                if not silent:
                    warn(msg)
                elif debug:
                    dprint(f"[{self}]\n" + str(msg))
                result = None
            if result is None and break_on_error:
                if rollback:
                    session.rollback()
                break
            elif result is not None and hook is not None:
                hook_queries = hook(session)
                if hook_queries:
                    hook_results = self.exec_queries(
                        hook_queries,
                        break_on_error = break_on_error,
                        rollback = rollback,
                        silent = silent,
                        debug = debug,
                    )
                    result = (result, hook_results)

            results.append(result)

    return results


def to_sql(
        self,
        df: pandas.DataFrame,
        name: str = None,
        index: bool = False,
        if_exists: str = 'replace',
        method: str = "",
        chunksize: Optional[int] = -1,
        schema: Optional[str] = None,
        silent: bool = False,
        debug: bool = False,
        as_tuple: bool = False,
        as_dict: bool = False,
        **kw
    ) -> Union[bool, SuccessTuple]:
    """
    Upload a DataFrame's contents to the SQL server.

    Parameters
    ----------
    df: pd.DataFrame
        The DataFrame to be uploaded.

    name: str
        The name of the table to be created.

    index: bool, default False
        If True, creates the DataFrame's indices as columns.

    if_exists: str, default 'replace'
        Drop and create the table ('replace') or append if it exists
        ('append') or raise Exception ('fail').
        Options are ['replace', 'append', 'fail'].

    method: str, default ''
        None or multi. Details on pandas.to_sql.

    chunksize: Optional[int], default -1
        How many rows to insert at a time.

    schema: Optional[str], default None
        Optionally override the schema for the table.
        Defaults to `SQLConnector.schema`.

    as_tuple: bool, default False
        If `True`, return a (success_bool, message) tuple instead of a `bool`.
        Defaults to `False`.

    as_dict: bool, default False
        If `True`, return a dictionary of transaction information.
        The keys are `success`, `msg`, `start`, `end`, `duration`, `num_rows`, `chunksize`,
        `method`, and `target`.
        
    kw: Any
        Additional arguments will be passed to the DataFrame's `to_sql` function

    Returns
    -------
    Either a `bool` or a `SuccessTuple` (depends on `as_tuple`).
    """
    import time
    import json
    import decimal
    from decimal import Decimal, Context
    from meerschaum.utils.warnings import error, warn
    import warnings
    import functools
    if name is None:
        error(f"Name must not be `None` to insert data into {self}.")

    ### We're requiring `name` to be positional, and sometimes it's passed in from background jobs.
    kw.pop('name', None)

    schema = schema or self.schema

    from meerschaum.utils.sql import (
        sql_item_name,
        table_exists,
        json_flavors,
        truncate_item_name,
    )
    from meerschaum.utils.dataframe import get_json_cols, get_numeric_cols
    from meerschaum.utils.dtypes import are_dtypes_equal, quantize_decimal
    from meerschaum.utils.dtypes.sql import NUMERIC_PRECISION_FLAVORS
    from meerschaum.connectors.sql._create_engine import flavor_configs
    from meerschaum.utils.packages import attempt_import, import_pandas
    sqlalchemy = attempt_import('sqlalchemy', debug=debug)
    pd = import_pandas()
    is_dask = 'dask' in df.__module__

    stats = {'target': name, }
    ### resort to defaults if None
    if method == "":
        if self.flavor in _bulk_flavors:
            method = functools.partial(psql_insert_copy, schema=self.schema)
        else:
            ### Should resolve to 'multi' or `None`.
            method = flavor_configs.get(self.flavor, {}).get('to_sql', {}).get('method', 'multi')
    stats['method'] = method.__name__ if hasattr(method, '__name__') else str(method)

    default_chunksize = self._sys_config.get('chunksize', None)
    chunksize = chunksize if chunksize != -1 else default_chunksize
    if chunksize is not None and self.flavor in _max_chunks_flavors:
        if chunksize > _max_chunks_flavors[self.flavor]:
            if chunksize != default_chunksize:
                warn(
                    f"The specified chunksize of {chunksize} exceeds the maximum of "
                    + f"{_max_chunks_flavors[self.flavor]} for flavor '{self.flavor}'.\n"
                    + f"    Falling back to a chunksize of {_max_chunks_flavors[self.flavor]}.",
                    stacklevel = 3,
                )
            chunksize = _max_chunks_flavors[self.flavor]
    stats['chunksize'] = chunksize

    success, msg = False, "Default to_sql message"
    start = time.perf_counter()
    if debug:
        msg = f"[{self}] Inserting {len(df)} rows with chunksize: {chunksize}..."
        print(msg, end="", flush=True)
    stats['num_rows'] = len(df)

    ### Check if the name is too long.
    truncated_name = truncate_item_name(name, self.flavor)
    if name != truncated_name:
        warn(
            f"Table '{name}' is too long for '{self.flavor}',"
            + f" will instead create the table '{truncated_name}'."
        )

    ### filter out non-pandas args
    import inspect
    to_sql_params = inspect.signature(df.to_sql).parameters
    to_sql_kw = {}
    for k, v in kw.items():
        if k in to_sql_params:
            to_sql_kw[k] = v

    to_sql_kw.update({
        'name': truncated_name,
        'schema': schema,
        ('con' if not is_dask else 'uri'): (self.engine if not is_dask else self.URI),
        'index': index,
        'if_exists': if_exists,
        'method': method,
        'chunksize': chunksize,
    })
    if is_dask:
        to_sql_kw.update({
            'parallel': True,
        })

    if self.flavor == 'oracle':
        ### For some reason 'replace' doesn't work properly in pandas,
        ### so try dropping first.
        if if_exists == 'replace' and table_exists(name, self, schema=schema, debug=debug):
            success = self.exec(
                "DROP TABLE " + sql_item_name(name, 'oracle', schema)
            ) is not None
            if not success:
                warn(f"Unable to drop {name}")


        ### Enforce NVARCHAR(2000) as text instead of CLOB.
        dtype = to_sql_kw.get('dtype', {})
        for col, typ in df.dtypes.items():
            if are_dtypes_equal(str(typ), 'object'):
                dtype[col] = sqlalchemy.types.NVARCHAR(2000)
            elif are_dtypes_equal(str(typ), 'int'):
                dtype[col] = sqlalchemy.types.INTEGER
        to_sql_kw['dtype'] = dtype
    elif self.flavor == 'mssql':
        dtype = to_sql_kw.get('dtype', {})
        for col, typ in df.dtypes.items():
            if are_dtypes_equal(str(typ), 'bool'):
                dtype[col] = sqlalchemy.types.INTEGER
        to_sql_kw['dtype'] = dtype

    ### Check for JSON columns.
    if self.flavor not in json_flavors:
        json_cols = get_json_cols(df)
        if json_cols:
            for col in json_cols:
                df[col] = df[col].apply(
                    (
                        lambda x: json.dumps(x, default=str, sort_keys=True)
                        if not isinstance(x, Hashable)
                        else x
                    )
                )

    ### Check for numeric columns.
    numeric_scale, numeric_precision = NUMERIC_PRECISION_FLAVORS.get(self.flavor, (None, None))
    if numeric_precision is not None and numeric_scale is not None:
        numeric_cols = get_numeric_cols(df)
        for col in numeric_cols:
            df[col] = df[col].apply(
                lambda x: (
                    quantize_decimal(x, numeric_scale, numeric_precision)
                    if isinstance(x, Decimal)
                    else x
                )
            )

    try:
        with warnings.catch_warnings():
            warnings.filterwarnings('ignore', 'case sensitivity issues')
            df.to_sql(**to_sql_kw)
        success = True
    except Exception as e:
        if not silent:
            warn(str(e))
        success, msg = False, str(e)

    end = time.perf_counter()
    if success:
        msg = f"It took {round(end - start, 2)} seconds to sync {len(df)} rows to {name}."
    stats['start'] = start
    stats['end'] = end
    stats['duration'] = end - start

    if debug:
        print(f" done.", flush=True)
        dprint(msg)

    stats['success'] = success
    stats['msg'] = msg
    if as_tuple:
        return success, msg
    if as_dict:
        return stats
    return success


def psql_insert_copy(
        table: pandas.io.sql.SQLTable,
        conn: Union[sqlalchemy.engine.Engine, sqlalchemy.engine.Connection],
        keys: List[str],
        data_iter: Iterable[Any],
        schema: Optional[str] = None,
    ) -> None:
    """
    Execute SQL statement inserting data for PostgreSQL.

    Parameters
    ----------
    table: pandas.io.sql.SQLTable
    
    conn: Union[sqlalchemy.engine.Engine, sqlalchemy.engine.Connection]
    
    keys: List[str]
        Column names
    
    data_iter: Iterable[Any]
        Iterable that iterates the values to be inserted

    schema: Optional[str], default None
        Optionally specify the schema of the table to be inserted into.

    Returns
    -------
    None
    """
    import csv
    from io import StringIO
    import json

    from meerschaum.utils.sql import sql_item_name

    ### NOTE: PostgreSQL doesn't support NUL chars in text, so they're removed from strings.
    data_iter = (
        (
            (
                (
                    json.dumps(item).replace('\0', '').replace('\\u0000', '')
                    if isinstance(item, (dict, list))
                    else (
                        item
                        if not isinstance(item, str)
                        else item.replace('\0', '').replace('\\u0000', '')
                    )
                )
            ) if item is not None
            else r'\N'
            for item in row
        ) for row in data_iter
    )

    dbapi_conn = conn.connection
    with dbapi_conn.cursor() as cur:
        s_buf = StringIO()
        writer = csv.writer(s_buf)
        writer.writerows(data_iter)
        s_buf.seek(0)

        columns = ', '.join(f'"{k}"' for k in keys)
        table_name = sql_item_name(table.name, 'postgresql', table.schema)
        sql = f"COPY {table_name} ({columns}) FROM STDIN WITH CSV NULL '\\N'"
        cur.copy_expert(sql=sql, file=s_buf)


def format_sql_query_for_dask(query: str) -> 'sqlalchemy.sql.selectable.Select':
    """
    Given a `SELECT` query, return a `sqlalchemy` query for Dask to use.
    This may only work with specific database flavors like PostgreSQL.

    Parameters
    ----------
    query: str
        The `SELECT` query to be parsed.

    Returns
    -------
    A `sqlalchemy` selectable.
    """
    if 'sqlalchemy' in str(type(query)):
        return query

    if 'select ' not in query.lower():
        raise ValueError(f"Cannot convert query to SQLAlchemy:\n\n{query}")

    def _remove_leading_select(q: str) -> str:
        return q.replace("SELECT ", "", 1)

    from meerschaum.utils.packages import attempt_import
    sqlalchemy_sql = attempt_import("sqlalchemy.sql")
    select, text = sqlalchemy_sql.select, sqlalchemy_sql.text

    parts = query.rsplit('ORDER BY', maxsplit=1)
    meta_query = f"SELECT * FROM (\n{query}\n) AS s"
    #  if parts[1]:
        #  meta_query += "\nORDER BY " + parts[1]
    return select(text(_remove_leading_select(meta_query)))



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Log into API connectors.
"""

from __future__ import annotations
from meerschaum.utils.typing import SuccessTuple, Optional, List, Any

def login(
        action: Optional[List[str]] = None,
        connector_keys: Optional[List[str]] = None,
        yes: bool = False,
        noask: bool = False,
        debug: bool = False,
        **kw: Any
    ) -> SuccessTuple:
    """
    Log into a Meerschaum API instance.
    """
    from meerschaum.utils.prompt import prompt, get_password, yes_no
    from meerschaum.connectors.parse import parse_instance_keys
    from meerschaum.config import get_config
    from meerschaum.config._edit import write_config
    from meerschaum.utils.warnings import warn
    from meerschaum.utils.formatting._shell import clear_screen
    from meerschaum.utils.formatting import print_tuple
    if action is None:
        action = []
    if connector_keys is None:
        connector_keys = []
    _possible_keys = action + connector_keys
    _keys = []
    for k in _possible_keys:
        if not k.startswith('api:'):
            warn(f"Connector '{k}' is not an API connector. Skipping...", stack=False)
            continue
        _keys.append(k)

    def _get_creds(connector):
        username = prompt(f"Username for connector '{connector}':")
        password = get_password(username=username, confirm=False)
        return username, password

    def _forget_creds(_dict):
        to_delete = ['username', 'password']
        for cred in to_delete:
            try:
                del _dict[cred]
            except KeyError:
                pass

    _connectors = []
    for k in _keys:
        try:
            _connectors.append(parse_instance_keys(k))
        except Exception as e:
            warn(f"Unable to build connector '{k}'. Is it registered?", stack=False)

    meerschaum_config = get_config('meerschaum')
    successes = 0
    for c in _connectors:
        while True:
            login_change = False
            if 'username' not in c.__dict__ or 'password' not in c.__dict__:
                login_change = True
                try:
                    clear_screen(debug=debug)
                    username, password = _get_creds(c)
                except KeyboardInterrupt:
                    return False, "Cancelled login."
                else:
                    c.username = username
                    c.password = password
            try:
                login_tuple = c.login(debug=debug)
            except Exception as e:
                warn(f"Failed to log in to connector '{c}':\n\n{e}\n\nSkipping '{c}'...", stack=False)
                break
            if not login_tuple[0]:
                if yes_no(
                    f"Would you like to try different login credentials for connector '{c}'?",
                    yes=yes, noask=noask
                ):
                    _forget_creds(c.__dict__)
                    continue
                else:
                    ### Did not have prior creds, was unsuccessful, and wants to abort.
                    if login_change:
                        _forget_creds(c.__dict__)
            print_tuple(login_tuple)
            successes += (1 if login_tuple[0] else 0)
            if login_tuple[0] and login_change:
                if yes_no(
                    f"Would you like to save the new credentials for connector '{c}'?",
                    yes=yes, noask=noask
                ):
                    meerschaum_config['connectors'][c.type][c.label].update(
                        {'username' : c.username, 'password' : c.password}
                    )
                    write_config({'meerschaum' : meerschaum_config})
                else:
                    _forget_creds(c.__dict__)
            break
        
    msg = f"Logged into {successes} connector" + ('s' if successes != 1 else '') + '.'

    return successes > 0, msg

def _complete_login(action: Optional[List[str]] = None, **kw: Any) -> List[str]:
    from meerschaum.utils.misc import get_connector_labels
    search_term = action[-1] if action else ''
    return get_connector_labels('api', search_term=search_term)



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Enforce data types for a pipe's underlying table.
"""

from __future__ import annotations
from io import StringIO
from meerschaum.utils.typing import Dict, Any, Optional

def enforce_dtypes(
        self,
        df: 'pd.DataFrame',
        chunksize: Optional[int] = -1,
        safe_copy: bool = True,
        debug: bool = False,
    ) -> 'pd.DataFrame':
    """
    Cast the input dataframe to the pipe's registered data types.
    If the pipe does not exist and dtypes are not set, return the dataframe.
    """
    import traceback
    from meerschaum.utils.warnings import warn
    from meerschaum.utils.debug import dprint
    from meerschaum.utils.dataframe import parse_df_datetimes, enforce_dtypes as _enforce_dtypes
    from meerschaum.utils.packages import import_pandas
    pd = import_pandas(debug=debug)
    if df is None:
        if debug:
            dprint(
                f"Received None instead of a DataFrame.\n"
                + "    Skipping dtype enforcement..."
            )
        return df

    pipe_dtypes = self.dtypes

    try:
        if isinstance(df, str):
            df = parse_df_datetimes(
                pd.read_json(StringIO(df)),
                ignore_cols = [
                    col
                    for col, dtype in pipe_dtypes.items()
                    if 'datetime' not in str(dtype)
                ],
                chunksize = chunksize,
                debug = debug,
            )
        else:
            df = parse_df_datetimes(
                df,
                ignore_cols = [
                    col
                    for col, dtype in pipe_dtypes.items()
                    if 'datetime' not in str(dtype)
                ],
                chunksize = chunksize,
                debug = debug,
            )
    except Exception as e:
        warn(f"Unable to cast incoming data as a DataFrame...:\n{e}\n\n{traceback.format_exc()}")
        return None

    if not pipe_dtypes:
        if debug:
            dprint(
                f"Could not find dtypes for {self}.\n"
                + "    Skipping dtype enforcement..."
            )
        return df

    return _enforce_dtypes(df, pipe_dtypes, safe_copy=safe_copy, debug=debug)


def infer_dtypes(self, persist: bool=False, debug: bool=False) -> Dict[str, Any]:
    """
    If `dtypes` is not set in `meerschaum.Pipe.parameters`,
    infer the data types from the underlying table if it exists.

    Parameters
    ----------
    persist: bool, default False
        If `True`, persist the inferred data types to `meerschaum.Pipe.parameters`.

    Returns
    -------
    A dictionary of strings containing the pandas data types for this Pipe.
    """
    if not self.exists(debug=debug):
        dtypes = {}
        if not self.columns:
            return {}
        dt_col = self.columns.get('datetime', None)
        if dt_col:
            if not self.parameters.get('dtypes', {}).get(dt_col, None):
                dtypes[dt_col] = 'datetime64[ns]'
        return dtypes

    from meerschaum.utils.sql import get_pd_type
    from meerschaum.utils.misc import to_pandas_dtype
    columns_types = self.get_columns_types(debug=debug)

    ### NOTE: get_columns_types() may return either the types as
    ###       PostgreSQL- or Pandas-style.
    dtypes = {
        c: (
            get_pd_type(t, allow_custom_dtypes=True)
            if str(t).isupper()
            else to_pandas_dtype(t)
        )
        for c, t in columns_types.items()
    } if columns_types else {}
    if persist:
        self.dtypes = dtypes
        self.edit(interactive=False, debug=debug)
    return dtypes



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Schedule processes and threads.
"""

from __future__ import annotations
from meerschaum.utils.typing import Callable, Any, Optional

def schedule_function(
        function: Callable[[Any], Any],
        frequency: str,
        *args,
        debug: bool = False,
        **kw
    ) -> None:
    """
    Block the process and execute the function intermittently according to the frequency.
    https://rocketry.readthedocs.io/en/stable/condition_syntax/index.html

    Parameters
    ----------
    function: Callable[[Any], Any]
        The function to execute.

    frequency: str
        The frequency at which `function` should be executed (e.g. `'daily'`).

    """
    import warnings
    from meerschaum.utils.warnings import warn
    from meerschaum.utils.packages import attempt_import
    from meerschaum.utils.misc import filter_keywords
    from concurrent.futures._base import CancelledError
    kw['debug'] = debug
    kw = filter_keywords(function, **kw)

    def _wrapper():
        return function(*args, **kw)

    pydantic = attempt_import('pydantic', debug=debug, lazy=False)
    rocketry = attempt_import('rocketry', debug=debug, lazy=False)
    try:
        app = rocketry.Rocketry()
        FuncTask = rocketry.tasks.FuncTask
        with warnings.catch_warnings():
            warnings.filterwarnings('ignore', 'Task\'s session not defined.')
            task = FuncTask(_wrapper, start_cond=frequency)
            app.session.add_task(task)
        return app.run(debug=debug)
    except (KeyboardInterrupt, CancelledError):
        try:
            app.session.shut_down(force=True)
        except CancelledError:
            pass
        return None
    except AttributeError:
        warn(
            "Failed to import scheduler.\n\n   "
            + "Run `mrsm install package 'pydantic<2.0.0'` and try again.",
            stack = False,
        )

